"""
statistic calculation for SRVF (curves) open and closed using Karcher
Mean and Variance

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
from numpy import (
    zeros,
    sqrt,
    fabs,
    cos,
    sin,
    tile,
    vstack,
    empty,
    cov,
    inf,
    mean,
    arange,
    prod,
)
from numpy import append, arccos, cumsum, sum, linspace
from numpy.linalg import svd
from numpy.random import randn
import fdasrsf.curve_functions as cf
import fdasrsf.utility_functions as uf
import fdasrsf.plot_style as plot
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
import collections


class fdacurve:
    """
    This class provides alignment methods for open and closed curves using the SRVF framework

    Usage: obj = fdacurve(beta, mode, N, scale)
    :param beta: numpy ndarray of shape (n, M, N) describing N curves
    in R^M
    :param mode: Open ('O') or closed curve ('C') (default 'O')
    :param N: resample curve to N points
    :param scale: scale curve to length 1 (true/false)
    :param q:           (n,T,K) matrix defining n dimensional srvf on T samples with K srvfs
    :param betan:       aligned curves
    :param qn:          aligned srvfs
    :param basis:       calculated basis
    :param beta_mean:   karcher mean curve
    :param q_mean:      karcher mean srvf
    :param gams:        warping functions
    :param v:           shooting vectors
    :param C:           karcher covariance
    :param s:           pca singular values
    :param U:           pca singular vectors
    :param coef:        pca coefficients
    :param pca          principal directions
    :param qun:         cost function
    :param lambda: lambda
    :param samples:     random samples
    :param gamr:        random warping functions
    :param cent:        center
    :param scale:       scale
    :param len:         length of curve
    :param len_q:       length of srvf
    :param mean_scale   mean length
    :param mean_scale_q mean length srvf
    :param E:           energy

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  26-Aug-2020
    """

    def __init__(self, beta, mode="O", N=200, scale=False):
        """
        fdacurve Construct an instance of this class
        :param beta: (n,T,K) matrix defining n dimensional curve on T samples with K curves
        :param mode:  Open ('O') or closed curve ('C') (default 'O')
        :param N: resample curve to N points
        :param scale: include scale (true/false)
        """
        self.mode = mode
        self.scale = scale

        K = beta.shape[2]
        n = beta.shape[0]
        q = zeros((n, N, K))
        beta1 = zeros((n, N, K))
        cent1 = zeros((n, K))
        len1 = zeros(K)
        lenq1 = zeros(K)
        for ii in range(0, K):
            if beta.shape[1] != N:
                beta1[:, :, ii] = cf.resamplecurve(beta[:, :, ii], N, mode=mode)
            else:
                beta1[:, :, ii] = beta[:, :, ii]
            a = -cf.calculatecentroid(beta1[:, :, ii])
            beta1[:, :, ii] += tile(a, (N, 1)).T
            q[:, :, ii], len1[ii], lenq1[ii] = cf.curve_to_q(beta1[:, :, ii], mode)
            cent1[:, ii] = -a

        self.q = q
        self.beta = beta1
        self.cent = cent1
        self.len = len1
        self.len_q = lenq1

    def karcher_mean(
        self, rotation=True, parallel=False, lam=0.0, cores=-1, method="DP"
    ):
        """
        This calculates the mean of a set of curves
        :param rotation: compute optimal rotation (default = T)
        :param parallel: run in parallel (default = F)
        :param lam: controls the elasticity (default = 0)
        :param cores: number of cores for parallel (default = -1 (all))
        :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"
        """
        n, T, N = self.beta.shape

        modes = ["O", "C"]
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        # Initialize mu as one of the shapes
        mu = self.q[:, :, 0]
        betamean = self.beta[:, :, 0]
        itr = 0

        gamma = zeros((T, N))
        maxit = 20

        sumd = zeros(maxit + 1)
        v = zeros((n, T, N))
        sumv = zeros((n, T))
        normvbar = zeros(maxit + 1)

        delta = 0.5
        tolv = 1e-4
        told = 5 * 1e-3

        print(
            "Computing Karcher Mean of %d curves in SRVF space with lam=%02f" % (N, lam)
        )
        while itr < maxit:
            print("updating step: %d" % (itr + 1))

            if iter == maxit:
                print("maximal number of iterations reached")

            mu = mu / sqrt(cf.innerprod_q2(mu, mu))
            if mode == 1:
                self.basis = cf.find_basis_normal(mu)
            else:
                self.basis = []

            sumv = zeros((n, T))
            sumd[0] = inf
            sumd[itr + 1] = 0
            out = Parallel(n_jobs=cores)(
                delayed(karcher_calc)(
                    mu, self.q[:, :, n], self.basis, mode, lam, rotation, method
                )
                for n in range(N)
            )
            v = zeros((n, T, N))
            gamma = zeros((T, N))
            for i in range(0, N):
                v[:, :, i] = out[i][0]
                gamma[:, i] = out[i][1]
                sumv += v[:, :, i]
                sumd[itr + 1] = sumd[itr + 1] + out[i][2] ** 2

            sumv = v.sum(axis=2)

            # Compute average direction of tangent vectors v_i
            vbar = sumv / float(N)

            normvbar[itr] = sqrt(cf.innerprod_q2(vbar, vbar))
            normv = normvbar[itr]

            if (sumd[itr] - sumd[itr + 1]) < 0:
                break
            elif normv > tolv and fabs(sumd[itr + 1] - sumd[itr]) > told:
                # Update mu in direction of vbar
                mu = (
                    cos(delta * normvbar[itr]) * mu
                    + sin(delta * normvbar[itr]) * vbar / normvbar[itr]
                )

                if mode == 1:
                    mu = cf.project_curve(mu)

                x = cf.q_to_curve(mu)
                a = -1 * cf.calculatecentroid(x)
                betamean = x + tile(a, [T, 1]).T
            else:
                break

            itr += 1

        # compute average length
        if self.scale:
            self.mean_scale = (prod(self.len)) ** (1.0 / self.len.shape[0])
            self.mean_scale_q = (prod(self.len_q)) ** (1.0 / self.len.shape[0])
            betamean = self.mean_scale * betamean

        self.q_mean = mu
        self.beta_mean = betamean
        self.v = v
        self.qun = sumd[0 : (itr + 1)]
        self.E = normvbar[0 : (itr + 1)]

        return

    def srvf_align(self, rotation=True, lam=0.0, parallel=False, cores=-1, method="DP"):
        """
        This aligns a set of curves to the mean and computes mean if not computed
        :param rotation: compute optimal rotation (default = T)
        :param lam: controls the elasticity (default = 0)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"
        """
        n, T, N = self.beta.shape

        modes = ["O", "C"]
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        # find mean
        if not hasattr(self, "beta_mean"):
            self.karcher_mean(
                rotation=rotation,
                lam=lam,
                parallel=parallel,
                cores=cores,
                method=method,
            )

        self.qn = zeros((n, T, N))
        self.betan = zeros((n, T, N))
        self.gams = zeros((T, N))
        centroid2 = cf.calculatecentroid(self.beta_mean)
        self.beta_mean = self.beta_mean - tile(centroid2, [T, 1]).T

        # align to mean
        out = Parallel(n_jobs=-1)(
            delayed(cf.find_rotation_and_seed_unique)(
                self.q_mean, self.q[:, :, n], mode, lam, rotation, method
            )
            for n in range(N)
        )
        for ii in range(0, N):
            self.gams[:, ii] = out[ii][2]
            self.qn[:, :, ii] = out[ii][0]
            btmp = out[ii][1].dot(self.beta[:, :, ii])
            self.betan[:, :, ii] = cf.group_action_by_gamma_coord(btmp, out[ii][2])

        return

    def karcher_cov(self):
        """
        This calculates the mean of a set of curves

        """
        if not hasattr(self, "beta_mean"):
            self.karcher_mean()

        M, N, K = self.v.shape

        N1 = M * N
        if self.scale:
            N1 += 1

        tmpv = zeros((N1, K))
        for i in range(0, K):
            tmp = self.v[:, :, i]
            tmpv1 = tmp.flatten()
            if self.scale:
                tmpv1 = append(tmpv1, self.len[i])

            tmpv[:, i] = tmpv1

        if self.scale:
            VM = mean(self.v, 2)
            VM = VM.flatten()
            VM = append(VM, self.mean_scale)
            tmpv = tmpv - tile(VM, [K, 1]).T
            self.C = cov(tmpv)
        else:
            self.C = cov(tmpv)

        return

    def shape_pca(self, no=10):
        """
        Computes principal direction of variation specified by no. N is
        Number of shapes away from mean. Creates 2*N+1 shape sequence

        :param no: number of direction (default 3)
        """
        if not hasattr(self, "C"):
            self.karcher_cov()

        U1, s, V = svd(self.C)
        self.U = U1[:, 0:no]
        self.s = s[0:no]

        # express shapes as coefficients
        K = self.beta.shape[2]
        VM = mean(self.v, 2)
        VM = VM.flatten()
        if self.scale:
            VM = append(VM, self.mean_scale)

        x = zeros((no, K))
        for ii in range(0, K):
            tmpv = self.v[:, :, ii]
            tmpv1 = tmpv.flatten()
            if self.scale:
                tmpv1 = append(tmpv1, self.len[ii])
            Utmp = self.U.T
            x[:, ii] = Utmp.dot((tmpv1 - VM))

        self.coef = x

        modes = ["O", "C"]
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        n1, T, N1 = self.beta.shape
        p = zeros((n1, T, no, 10))
        for j in range(0, no):
            for i in range(1, 11):
                tmp = VM + 0.5 * (i - 5) * sqrt(self.s[j]) * self.U[:, j]
                m, n = self.q_mean.shape
                if self.scale:
                    tmp_scale = tmp[-1]
                    tmp = tmp[0:-1]
                else:
                    tmp_scale = 1
                v1 = tmp.reshape(m, n)
                q2n = cf.elastic_shooting(self.q_mean, v1, mode)

                p[:, :, j, i - 1] = cf.q_to_curve(q2n, tmp_scale)

        self.pca = p

        return

    def sample_shapes(self, no=3, numSamp=10):
        """
        Computes sample shapes from mean and covariance

        :param no: number of direction (default 3)
        :param numSamp: number of samples (default 10)
        """
        n, T = self.q_mean.shape
        modes = ["O", "C"]
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        U, s, V = svd(self.C)

        if mode == 0:
            N = 2
        else:
            N = 10

        epsilon = 1.0 / (N - 1)

        samples = empty(numSamp, dtype=object)
        for i in range(0, numSamp):
            v = zeros((2, T))
            for m in range(0, no):
                v = v + randn() * sqrt(s[m]) * vstack((U[0:T, m], U[T : 2 * T, m]))

            q1 = self.q_mean
            for j in range(0, N - 1):
                normv = sqrt(cf.innerprod_q2(v, v))

                if normv < 1e-4:
                    q2 = self.q_mean
                else:
                    q2 = cos(epsilon * normv) * q1 + sin(epsilon * normv) * v / normv
                    if mode == 1:
                        q2 = cf.project_curve(q2)

                # Parallel translate tangent vector
                basis2 = cf.find_basis_normal(q2)
                v = cf.parallel_translate(v, q1, q2, basis2, mode)

                q1 = q2

            samples[i] = cf.q_to_curve(q2)

        self.samples = samples
        return

    def plot(self, multivariate=False):
        """
        plot curve mean results

        :param multivariate: plot as multivariate functions instead of curves (default=False)
        """
        if multivariate:
            n, T, K = self.beta.shape
            for jj in range(n):
                fig, ax = plt.subplots()
                for ii in range(0, K):
                    ax.plot(linspace(0, 1, T), self.beta[jj, :, ii])
                plt.title("Original Function: %d" % (jj + 1))

        else:
            fig, ax = plt.subplots()
            n, T, K = self.beta.shape
            for ii in range(0, K):
                ax.plot(self.beta[0, :, ii], self.beta[1, :, ii])
            plt.title("Curves")
            ax.set_aspect("equal")
            plt.axis("off")
            plt.gca().invert_yaxis()

        if hasattr(self, "gams"):
            M = self.gams.shape[0]
            fig, ax = plot.f_plot(
                arange(0, M) / float(M - 1), self.gams, title="Warping Functions"
            )

        if hasattr(self, "beta_mean"):
            if multivariate:
                for jj in range(n):
                    fig, ax = plt.subplots()
                    ax.plot(linspace(0, 1, T), self.beta_mean[jj, :])
                    plt.title("Karcher Mean: %d" % (jj + 1))
            else:
                fig, ax = plt.subplots()
                ax.plot(self.beta_mean[0, :], self.beta_mean[1, :])
                plt.title("Karcher Mean")
                ax.set_aspect("equal")
                plt.axis("off")
                plt.gca().invert_yaxis()

        if multivariate:
            if hasattr(self, "betan"):
                n, T, K = self.beta.shape
                for jj in range(n):
                    fig, ax = plt.subplots()
                    for ii in range(0, K):
                        ax.plot(linspace(0, 1, T), self.betan[jj, :, ii])
                    plt.title("Aligned Function: %d" % (jj + 1))

        plt.show()

    def plot_pca(self):
        if not hasattr(self, "s"):
            raise NameError("Calculate PCA")

        fig, ax = plt.subplots()
        ax.plot(cumsum(self.s) / sum(self.s) * 100)
        plt.title("Variability Explained")
        plt.xlabel("PC")

        # plot principal modes of variability
        VM = mean(self.v, 2)
        VM = VM.flatten()
        if self.scale:
            VM = append(VM, self.mean_scale)

        modes = ["O", "C"]
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        for j in range(0, 4):
            fig, ax = plt.subplots()
            for i in range(1, 11):
                tmp = VM + 0.5 * (i - 5) * sqrt(self.s[j]) * self.U[:, j]
                m, n = self.q_mean.shape
                if self.scale:
                    tmp_scale = tmp[-1]
                    tmp = tmp[0:-1]
                else:
                    tmp_scale = 1
                v1 = tmp.reshape(m, n)
                q2n = cf.elastic_shooting(self.q_mean, v1, mode)

                p = cf.q_to_curve(q2n, tmp_scale)

                mv = 0.2
                if self.scale:
                    mv *= self.mean_scale

                if i == 5:
                    ax.plot(mv * i + p[0, :], p[1, :], "k", linewidth=2)
                else:
                    ax.plot(mv * i + p[0, :], p[1, :], linewidth=2)

            ax.set_aspect("equal")
            plt.axis("off")
            plt.title("PD %d" % (j + 1))

        plt.show()


def karcher_calc(mu, q, basis, closed, lam, rotation, method):
    # Compute shooting vector from mu to q_i
    qn_t, R, gamI = cf.find_rotation_and_seed_unique(
        mu, q, closed, lam, rotation, method
    )
    qn_t = qn_t / sqrt(cf.innerprod_q2(qn_t, qn_t))

    q1dotq2 = cf.innerprod_q2(mu, qn_t)

    if q1dotq2 > 1:
        q1dotq2 = 1

    d = arccos(q1dotq2)

    u = qn_t - q1dotq2 * mu
    normu = sqrt(cf.innerprod_q2(u, u))
    if normu > 1e-4:
        w = u * arccos(q1dotq2) / normu
    else:
        w = zeros(qn_t.shape)

    # Project to tangent space of manifold to obtain v_i
    if closed == 0:
        v = w
    else:
        v = cf.project_tangent(w, q, basis)

    return (v, gamI, d)



"""
Distance metrics for functions and curves in R^n for use with UMAP
(https://github.com/lmcinnes/umap)

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numba
from numba.core.typing import cffi_utils
from _DP import ffi, lib
import _DP
from numpy import (
    linspace,
    interp,
    zeros,
    diff,
    double,
    sqrt,
    arange,
    int64,
    trapz,
    ones,
)
from numpy import (
    ascontiguousarray,
    roll,
    eye,
    arccos,
)
from numpy import kron, floor, mean
from numpy.linalg import norm, svd, det, solve


DP = lib.DP
cffi_utils.register_module(_DP)


@numba.njit()
def grad(f, binsize):
    n = f.shape[0]
    g = zeros(n)
    h = binsize * arange(1, n + 1)
    g[0] = (f[1] - f[0]) / (h[1] - h[0])
    g[-1] = (f[-1] - f[(-2)]) / (h[-1] - h[-2])

    h = h[2:] - h[0:-2]
    g[1:-1] = (f[2:] - f[0:-2]) / h[0]

    return g


@numba.njit()
def warp(q1, q2):
    M = q1.shape[0]
    disp = 0
    n1 = 1
    lam = 0.0
    gam = zeros(M)
    q1 = q1 / norm(q1)
    q2 = q2 / norm(q2)
    q1 = ascontiguousarray(q1)
    q2 = ascontiguousarray(q2)
    gam = ascontiguousarray(gam)
    q2i = ffi.from_buffer(q2)
    q1i = ffi.from_buffer(q1)
    gami = ffi.from_buffer(gam)
    DP(q2i, q1i, n1, M, lam, disp, gami)

    return gam


@numba.njit()
def basis(q):
    n, T = q.shape
    e = eye(n)
    Ev = zeros((n, T, n))
    for i in range(0, n):
        x = e[:, i]
        Ev[:, :, i] = x.repeat(T).reshape((-1, T))

    qnorm = zeros(T)
    for t in range(0, T):
        qnorm[t] = norm(q[:, t])

    delG = list()
    for i in range(0, n):
        x = q[i, :] / qnorm
        tmp1 = x.repeat(n).reshape((-1, n)).T
        tmp2 = qnorm.repeat(n).reshape((-1, n)).T
        delG.append(tmp1 * q + tmp2 * Ev[:, :, i])

    return delG


@numba.njit()
def freshape(f):
    n, M = f.shape
    out1 = f.reshape(M * n)
    out = zeros(M * n)
    out[::2] = out1[0:M]
    out[1::2] = out1[M:]

    return out


@numba.njit()
def warp_curve(q1, q2):
    n1, M = q1.shape
    disp = 0
    lam = 0.0
    gam = zeros(M)
    q1i = freshape(q1)
    q2i = freshape(q2)

    q1i = ascontiguousarray(q1i)
    q2i = ascontiguousarray(q2i)

    gam = ascontiguousarray(gam)
    q1ptr = ffi.from_buffer(q1i)
    q2ptr = ffi.from_buffer(q2i)
    gami = ffi.from_buffer(gam)
    DP(q2ptr, q1ptr, n1, M, lam, disp, gami)

    return gam


@numba.njit()
def shift_curve(f, tau):
    n, T = f.shape
    fn = zeros((n, T))
    for i in range(n):
        fn[i, 0 : (T - 1)] = roll(f[i, 0 : (T - 1)], tau)
        fn[i, T - 1] = fn[i, 0]
    return fn


@numba.njit()
def c_to_q(beta, closed):
    n, T = beta.shape
    v = zeros((n, T))
    for i in range(n):
        v[i, :] = grad(beta[i, :], 1.0 / (T - 1))

    q = zeros((n, T))
    for i in range(T):
        L = sqrt(norm(v[:, i]))
        if L > 0.0001:
            q[:, i] = v[:, i] / L
        else:
            q[:, i] = v[:, i] * 0.0001

    tmp = q * q
    tmp.sum() / T
    tmp1 = tmp.sum() / T
    q = q / sqrt(tmp1)

    if closed == 1:
        q = proj_c(q)

    return q


@numba.njit()
def find_rot(q1, q2):
    q1 = ascontiguousarray(q1)
    q2 = ascontiguousarray(q2)
    n = q1.shape[0]
    A = q1.dot(q2.T)
    U, s, Vh = svd(A)

    tst = det(A)
    if tst > 0:
        S = eye(n)
    else:
        S = eye(n)
        S[:, -1] = -S[:, -1]

    R = U.dot(S).dot(Vh)
    return R


@numba.njit()
def inner_prod(q1, q2):
    T = q1.shape[1]
    cst = q1 * q2
    tmp = cst.sum() / T
    return tmp


@numba.njit()
def proj_c(q):
    n, T = q.shape
    if n == 2:
        dt = 0.35
    if n == 3:
        dt = 0.2
    epsilon = 1e-6

    iter = 1
    res = ones(n)
    J = zeros((n, n))

    s = linspace(0, 1, T)

    qnew = q / sqrt(inner_prod(q, q))
    qnorm = zeros(T)
    G = zeros(n)
    C = zeros(300)
    while norm(res) > epsilon:
        if iter > 300:
            break

        # Jacobian
        for i in range(0, n):
            for j in range(0, n):
                J[i, j] = 3 * trapz(qnew[i, :] * qnew[j, :], s)

        J += eye(n)

        for i in range(0, T):
            qnorm[i] = norm(qnew[:, i])

        # Compute the residue
        for i in range(0, n):
            G[i] = trapz(qnew[i, :] * qnorm, s)

        res = -G

        if norm(res) < epsilon:
            break

        x = solve(J, res)
        C[iter] = norm(res)

        delG = basis(qnew)
        temp = zeros((n, T))
        for i in range(0, n):
            temp += x[i] * delG[i] * dt

        qnew += temp
        iter += 1

    qnew = qnew / sqrt(inner_prod(qnew, qnew))

    return q


@numba.njit()
def find_seed_rot(beta1, beta2, closed):
    beta1 = ascontiguousarray(beta1)
    beta2 = ascontiguousarray(beta2)
    q1 = c_to_q(beta1, closed)

    n, T = beta1.shape
    scl = 4.0
    minE = 1000
    if closed == 1:
        end_idx = int(floor(T / scl))
        scli = 4
    else:
        end_idx = 0

    x = linspace(0, 1, T)
    for ctr in range(0, end_idx + 1):
        if closed == 1:
            shift = int(scli * ctr)
            beta2n = shift_curve(beta2, shift)
        else:
            beta2n = beta2

        R = find_rot(beta1, beta2n)
        beta2new = R.dot(beta2n)
        q2new = c_to_q(beta2new, closed)

        gam = warp_curve(q2new, q1)
        gamI = interp(x, gam, x)

        # apply warp
        beta_warp = zeros((n, T))
        for j in range(0, n):
            beta_warp[j, :] = interp(gamI, x, beta2new[j, :])

        q2new = c_to_q(beta_warp, closed)

        if closed == 1:
            q2new = proj_c(q2new)

        tmp = inner_prod(q1, q2new)
        if tmp > 1:
            tmp = 1
        Ec = arccos(tmp)
        if Ec < minE:
            q2best = q2new

    return q2best


@numba.njit()
def curve_center(beta):
    n, T = beta.shape
    betadot = zeros((n, T))
    for i in range(n):
        betadot[i, :] = grad(beta[i, :], 1.0 / (T - 1))

    normbetadot = zeros(T)
    integrand = zeros((n, T))
    for i in range(T):
        normbetadot[i] = norm(betadot[:, i])
        integrand[:, i] = beta[:, i] * normbetadot[i]

    scale = trapz(normbetadot, linspace(0, 1, T))
    centroid = zeros(n)
    for i in range(n):
        centroid[i] = trapz(integrand[i, :], linspace(0, 1, T)) / scale

    return centroid


@numba.njit()
def efda_distance(q1, q2, alpha=0):
    """ "
    calculates the distances between two curves, where
    q2 is aligned to q1. In other words calculates the elastic distances/
    This metric is set up for use with UMAP or t-sne from scikit-learn

    :param q1: vector of size N
    :param q2: vector of size N
    :param alpha: weight between phase and amplitude
                  (default = 0, returns amplitude)

    :rtype: scalar
    :return dist: amplitude distance
    """
    tst = q1 - q2
    if tst.sum() == 0:
        dist = 0
    else:
        q1 = q1.astype(double)
        q2 = q2.astype(double)
        gam = warp(q1, q2)
        M = q1.shape[0]
        time = linspace(0, 1, q1.shape[0])
        gam = (gam - gam[0]) / (gam[-1] - gam[0])
        gam_dev = grad(gam, 1 / double(M - 1))
        tmp = interp((time[-1] - time[0]) * gam + time[0], time, q2)

        qw = tmp * sqrt(gam_dev)

        y = (qw - q1) ** 2
        tmp = diff(time) * (y[0:-1] + y[1:]) / 2
        da = sqrt(tmp.sum())

        binsize = mean(diff(time))
        psi = sqrt(grad(gam, binsize))
        q1dotq2 = trapz(psi, time)
        if q1dotq2 > 1:
            q1dotq2 = 1
        elif q1dotq2 < -1:
            q1dotq2 = -1
        dp = arccos(q1dotq2)

        dist = (1 - alpha) * da + alpha * dp

    return dist


@numba.njit()
def efda_distance_curve(beta1, beta2, closed):
    """ "
    calculates the distances between two curves, where
    beta2 is aligned to beta1. In other words calculates the elastic distance.
    This metric is set up for use with UMAP or t-sne from scikit-learn

    :param beta1: vector of size n*M
    :param beta2: vector of size n*M
    :param closed: (0) if open curves and (1) if closed curves

    :rtype: scalar
    :return dist: shape distance
    """
    tst = beta1 - beta2
    if tst.sum() == 0:
        dist = 0.0
    else:
        n = int64(2)
        T = int64(beta1.shape[0] / n)
        beta1 = ascontiguousarray(beta1)
        beta2 = ascontiguousarray(beta2)
        beta1_i = beta1.reshape((n, T))
        beta2_i = beta2.reshape((n, T))
        beta1_i = beta1_i.astype(double)
        beta2_i = beta2_i.astype(double)

        centroid1 = curve_center(beta1_i)
        beta1_i = beta1_i - kron(ones((T, 1)), centroid1).T
        centroid2 = curve_center(beta2_i)
        beta2_i = beta2_i - kron(ones((T, 1)), centroid2).T

        q1 = c_to_q(beta1_i, closed)

        q1 = ascontiguousarray(q1)
        # optimize over SO(n) x Gamma
        q2 = find_seed_rot(beta1_i, beta2_i, closed)

        q1dotq2 = inner_prod(q1, q2)
        if q1dotq2 > 1:
            q1dotq2 = 1
        elif q1dotq2 < -1:
            q1dotq2 = -1
        dist = arccos(q1dotq2)

    return dist



# cython: language_level=2
from libcpp cimport bool

cdef extern from "UnitSquareImage.h":
    void findgrad2D(double *dfdu, double *dfdv, double *f, int n, int t, int d);
    int check_crossing(double *f, int n, int t, int D);



"""
functions for SRVF image manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
from scipy.interpolate import RegularGridInterpolator
import cimage as im


def apply_gam_imag(F, gam):
    """
    apply diffeomorphism to image

    :param F: numpy array of image
    :param gam: numpy array of diffeomorphism

    :rtype: np.array
    :return Fnew: warped image

    """
    ndim = F.ndim

    x2 = gam[:, :, 1].ravel(order="F")
    gamtmp = gam[:, :, 0].T
    y2 = np.sort(gamtmp.ravel(order="F"))

    if ndim == 3:
        (m, n, d) = F.shape
        Fnew = np.zeros((m, n, d))

        U = np.linspace(0, 1, m)
        V = np.linspace(0, 1, n)

        for j in range(d):
            interp = RegularGridInterpolator((U, V), F[:, :, j], method="linear")
            tmp = interp((x2, y2))
            Fnew[:, :, j] = tmp.reshape((m, n), order="F")
    elif ndim == 2:
        (m, n) = F.shape

        U = np.linspace(0, 1, m)
        V = np.linspace(0, 1, n)

        interp = RegularGridInterpolator((U, V), F, method="linear")
        tmp = interp((x2, y2))
        Fnew = tmp.reshape((m, n), order="F")

    return Fnew


def makediffeoid(nrow, ncol):
    """
    computes a 2-D identity diffeomorphism

    :param nrow: number of rows
    :param ncol: number of columns

    :rtype: np.array
    :return gamid: amplitude depth

    """
    D = 2
    gamid = np.zeros((nrow, ncol, D))

    U, V = np.meshgrid(np.linspace(0, 1, ncol), np.linspace(0, 1, nrow), indexing="xy")

    gamid[:, :, 0] = U
    gamid[:, :, 1] = V

    return gamid


def image_to_q(F):
    d = F.shape[2]

    if d < 2:
        raise NameError("Data dimension is wrong!")

    q = F.copy()

    sqrtmultfact = np.sqrt(Jacob_imag(F))
    for i in range(d):
        q[:, :, i] = sqrtmultfact * F[:, :, i]

    return q


def Jacob_imag(F):
    (m, n, d) = F.shape

    if d < 2:
        raise NameError("Data dimension is wrong!")

    F = np.ascontiguousarray(F)
    dfdu, dfdv = im.compgrad3D(F)
    dfdu = dfdu.reshape((m, n, d), order="F")
    dfdv = dfdv.reshape((m, n, d), order="F")

    multFactor = np.zeros((m, n))

    if d == 2:
        multFactor = dfdu[:, :, 0] * dfdv[:, :, 1] - dfdu[:, :, 1] * dfdv[:, :, 0]
        multFactor = np.abs(multFactor)
    elif d == 3:
        multFactor = (
            (dfdu[:, :, 1] * dfdv[:, :, 2] - dfdu[:, :, 2] * dfdv[:, :, 1]) ** 2
            + (dfdu[:, :, 0] * dfdv[:, :, 2] - dfdu[:, :, 2] * dfdv[:, :, 0]) ** 2
            + (dfdu[:, :, 0] * dfdv[:, :, 1] - dfdu[:, :, 1] * dfdv[:, :, 0]) ** 2
        )
        multFactor = np.sqrt(multFactor)

    return multFactor


def compEnergy(q1, q2):
    (m, n, d) = q1.shape
    ds = 1 / (m - 1) / (n - 1)

    tmp = q1 - q2

    H = np.dot(tmp.ravel(), tmp.ravel()) * ds

    return H


def updateGam(qt, qm, b):
    v = qt - qm
    w = findphistar(qt, b)
    gamupdate = findupdategam(v, w, b)

    return gamupdate


def findphistar(q, b):
    (m, n, D, K) = b.shape

    d = q.shape[2]
    dbxdu = np.zeros((m, n, K))
    dbydv = np.zeros((m, n, K))
    expr1 = np.zeros((m, n, d, K))
    expr2 = np.zeros((m, n, d, K))

    for k in range(K):
        btmp = b[:, :, 0, k]
        btmp = np.ascontiguousarray(btmp)
        dbxdui, tmp = im.compgrad2D(btmp)
        dbxdu[:, :, k] = dbxdui.reshape((m, n), order="F")
        btmp = b[:, :, 1, k]
        btmp = np.ascontiguousarray(btmp)
        tmp, dbydvi = im.compgrad2D(btmp)
        dbydv[:, :, k] = dbydvi.reshape((m, n), order="F")

    divb = dbxdu + dbydv

    dqdu, dqdv = im.compgrad3D(q)
    dqdu = dqdu.reshape(q.shape, order="F")
    dqdv = dqdv.reshape(q.shape, order="F")

    for k in range(K):
        for j in range(d):
            expr1[:, :, j, k] = divb[:, :, k] * q[:, :, j]
            expr2[:, :, j, k] = (
                dqdu[:, :, j] * b[:, :, 0, k] + dqdv[:, :, j] * b[:, :, 1, k]
            )

    w = 0.5 * expr1 + expr2

    return w


def findupdategam(v, w, b):
    (m, n, D, K) = b.shape
    ds = 1 / ((m - 1) * (n - 1))

    innp = np.zeros(K)

    gamupdate = np.zeros((m, n, D))

    for k in range(K):
        vt = w[:, :, :, k]
        innp[k] = np.dot(v.ravel(), vt.ravel()) * ds

        gamupdate = gamupdate + innp[k] * b[:, :, :, k]

    return gamupdate


def apply_gam_gamid(gamid, gaminc):
    (m, n, d) = gamid.shape

    U = np.linspace(0, 1, m)
    V = np.linspace(0, 1, n)

    gamcum = np.zeros((m, n, d))
    x2 = gaminc[:, :, 1].ravel(order="F")
    gamtmp = gaminc[:, :, 0].T
    y2 = np.sort(gamtmp.ravel(order="F"))
    for j in range(d):
        interp = RegularGridInterpolator((U, V), gamid[:, :, j], method="linear")
        tmp = interp((x2, y2))
        gamcum[:, :, j] = tmp.reshape((m, n), order="F")

    return gamcum


def gen_basis(m, n, M=10, N=10, baseType="t", ortho=False):
    b = formbasisTid(M, m, n, baseType)

    if ortho:
        b = GramSchmidt(b)

    return b


def formbasisTid(M, m, n, base_type):
    """
    Generate Basis on tangent space at identity

    :param M: number of basis functions
    :param m: number of rows
    :param n: number of colunns
    :param base_type: basis type ('t' only implemented)

    :rtype: np.array
    :return b: basis functions

    """
    U, V = np.meshgrid(np.linspace(0, 1, n), np.linspace(0, 1, m), indexing="xy")

    idx = 0

    if base_type == "t":
        b = np.zeros((m, n, 2, 2 * M))
        for s in range(1, M + 1):
            const = np.sqrt(2) * np.pi * s
            sPI2 = 2 * np.pi * s

            b[:, :, 0, idx] = np.zeros((m, n))
            b[:, :, 1, idx] = (np.cos(sPI2 * V) - 1) / const

            b[:, :, 0, idx + 1] = (np.cos(sPI2 * U) - 1) / const
            b[:, :, 1, idx + 1] = np.zeros((m, n))

            idx += 2

    return b


def GramSchmidt(b):
    (m, n, D, N) = b.shape

    ds = 1 / ((m - 1) * (n - 1))

    cnt = 0
    G = np.zeros((m, n, D, N))
    G[:, :, :, cnt] = b[:, :, :, cnt]

    Gtmp = G[:, :, :, cnt]
    Gtmp = np.ascontiguousarray(Gtmp)
    dvx, dvy = im.compgrad3D(Gtmp)
    dvx = dvx.reshape(Gtmp.shape, order="F")
    dvy = dvy.reshape(Gtmp.shape, order="F")
    l = np.dot(dvx.ravel(), dvx.ravel()) * ds + np.dot(dvy.ravel(), dvy.ravel()) * ds
    G[:, :, :, cnt] = G[:, :, :, cnt] / np.sqrt(l)

    for i in range(1, N):
        G[:, :, :, i] = b[:, :, :, i]
        Gtmp = G[:, :, :, i]
        Gtmp = np.ascontiguousarray(Gtmp)
        dv1x, dv1y = im.compgrad3D(Gtmp)
        dv1x = dv1x.reshape(Gtmp.shape, order="F")
        dv1y = dv1y.reshape(Gtmp.shape, order="F")

        for j in range(0, i - 1):
            Gtmp = G[:, :, :, j]
            Gtmp = np.ascontiguousarray(Gtmp)
            dv2x, dv2y = im.compgrad3D(Gtmp)
            dv2x = dv2x.reshape(Gtmp.shape, order="F")
            dv2y = dv2y.reshape(Gtmp.shape, order="F")
            t = (
                np.dot(dv1x.ravel(), dv2x.ravel()) * ds
                + np.dot(dv1y.ravel(), dv2y.ravel()) * ds
            )
            G[:, :, :, i] = G[:, :, :, i] - t * G[:, :, :, j]

        v = G[:, :, :, i]
        l = np.dot(v.ravel(), v.ravel()) * ds

        if l > 0:
            cnt += 1
            G[:, :, :, cnt] = G[:, :, :, i] / np.sqrt(l)

    return G



# cython: language_level=2
from libcpp cimport bool
cimport cyarma
from cyarma cimport vec

cdef extern from "rbfgs.h":
    vec rlbfgs_optim(vec q1, vec q2, vec time, int maxiter, double lam, int penalty);



# cython: language_level=2
import numpy as np
cimport crbfgs
cimport cyarma
cimport numpy as np

include "cyarma.pyx"

from libcpp cimport bool

def rlbfgs(np.ndarray[double, ndim=1, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] q2, 
           np.ndarray[double, ndim=1, mode="c"] time, maxiter=30, lam=0.0, penalty=0):
    q1 = np.ascontiguousarray(q1)
    q2 = np.ascontiguousarray(q2)
    time = np.ascontiguousarray(time)
    cdef vec aq1 = numpy_to_vec_d(q1)
    cdef vec aq2 = numpy_to_vec_d(q2)
    cdef vec atime = numpy_to_vec_d(time)
    cdef vec out = rlbfgs_optim(aq1, aq2, atime, maxiter, lam, penalty)

    cdef np.ndarray[np.double_t,ndim=1] gam = vec_to_numpy(out, None)

    return(gam)


def rlbfgs_dist(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] q2):
    q1 = np.ascontiguousarray(q1)
    q2 = np.ascontiguousarray(q2)
    d = np.zeros(q1.shape[1])
    M = q1.shape[0]
    alpha = 0.5
    time = np.linspace(0, 1, M)
    for i in range(q1.shape[1]):
        q1t = q1[:,1]
        q1t = np.ascontiguousarray(q1t)
        gam = rlbfgs(q1t, time, q2)
        # warp q
        gam_dev = np.gradient(gam, 1.0 / (M - 1))
        tmp = np.interp((time[-1] - time[0]) * gam + time[0], time, q2)

        qw = tmp * np.sqrt(gam_dev)
        Dy = np.sqrt(np.trapz((qw - q1t) ** 2, time))

        binsize = np.mean(np.diff(time))
        psi = np.sqrt(np.gradient(gam, binsize))
        q1dotq2 = np.trapz(psi, time)
        if q1dotq2 > 1:
            q1dotq2 = 1
        elif q1dotq2 < -1:
            q1dotq2 = -1

        Dx = np.real(np.arccos(q1dotq2))

        d[i] = alpha * Dy + (1-alpha) * Dx
    
    return d
        
        
    


"""
Utility functions for SRSF Manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from scipy.interpolate import UnivariateSpline, interp1d
from scipy.integrate import trapz, cumtrapz
from scipy.linalg import norm, svd, cholesky, inv, pinv
from scipy.stats.mstats import mquantiles
from scipy import special as sp
from numpy import zeros, interp, finfo, double, sqrt, diff, linspace
from numpy import arccos, sin, cos, arange, ascontiguousarray, round
from numpy import ones, real, pi, cumsum, fabs, cov, diagflat, inner
from numpy import gradient, column_stack, append, mean, hstack, median
from numpy import insert, vectorize, ceil, mod, array, percentile, dot
from numpy import exp
from joblib import Parallel, delayed
import numpy.random as rn
import optimum_reparamN2 as orN2
import optimum_reparam_N as orN
import cbayesian as bay
import fdasrsf.geometry as geo
from fdasrsf.rbfgs import rlbfgs
import sys


def smooth_data(f, sparam=1):
    """
    This function smooths a collection of functions using a box filter

    :param f: numpy ndarray of shape (M,N) of M functions with N samples
    :param sparam: Number of times to run box filter (default = 25)

    :rtype: numpy ndarray
    :return f: smoothed functions functions

    """
    M = f.shape[0]
    N = f.shape[1]

    fo = f.copy()
    for k in range(0, sparam):
        for r in range(0, N):
            fo[1 : (M - 2), r] = (
                fo[0 : (M - 3), r] + 2 * fo[1 : (M - 2), r] + fo[2 : (M - 1), r]
            ) / 4
    return fo


def gradient_spline(time, f, smooth=False):
    """
    This function takes the gradient of f using b-spline smoothing

    :param time: vector of size N describing the sample points
    :param f: numpy ndarray of shape (M,N) of M functions with N samples
    :param smooth: smooth data (default = F)

    :rtype: tuple of numpy ndarray
    :return f0: smoothed functions functions
    :return g: first derivative of each function
    :return g2: second derivative of each function

    """
    M = f.shape[0]

    if f.ndim > 1:
        N = f.shape[1]
        f0 = zeros((M, N))
        g = zeros((M, N))
        g2 = zeros((M, N))
        for k in range(0, N):
            if smooth:
                spar = time.shape[0] * (0.025 * fabs(f[:, k]).max()) ** 2
            else:
                spar = 0
            tmp_spline = UnivariateSpline(time, f[:, k], s=spar)
            f0[:, k] = tmp_spline(time)
            g[:, k] = tmp_spline(time, 1)
            g2[:, k] = tmp_spline(time, 2)
    else:
        if smooth:
            spar = time.shape[0] * (0.025 * fabs(f).max()) ** 2
        else:
            spar = 0
        tmp_spline = UnivariateSpline(time, f, s=spar)
        f0 = tmp_spline(time)
        g = tmp_spline(time, 1)
        g2 = tmp_spline(time, 2)

    return f0, g, g2


def f_to_srsf(f, time, smooth=False):
    """
    converts f to a square-root slope function (SRSF)

    :param f: vector of size N samples
    :param time: vector of size N describing the sample points

    :rtype: vector
    :return q: srsf of f

    """
    eps = finfo(double).eps
    f0, g, g2 = gradient_spline(time, f, smooth)
    q = g / sqrt(fabs(g) + eps)
    return q


def srsf_to_f(q, time, f0=0.0):
    """
    converts q (srsf) to a function

    :param q: vector of size N samples of srsf
    :param time: vector of size N describing time sample points
    :param f0: initial value

    :rtype: vector
    :return f: function

    """
    integrand = q * fabs(q)
    f = f0 + cumtrapz(integrand, time, initial=0)
    return f


def optimum_reparam(
    q1, time, q2, method="DP2", lam=0.0, penalty="roughness", grid_dim=7
):
    """
    calculates the warping to align srsf q2 to q1

    :param q1: vector of size N or array of NxM samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q2: vector of size N or array of NxM samples samples of second SRSF
    :param method: method to apply optimization (default="DP2") options are
                   "DP","DP2","RBFGS","cRBFGS"
    :param lam: controls the amount of elasticity (default = 0.0)
    :param penalty: penalty type (default="roughness") options are "roughness",
                    "l2gam", "l2psi", "geodesic". Only roughness implemented
                    in all methods. To use others method needs to be "RBFGS"
                    or "cRBFGS"
    :param grid_dim: size of the grid, for the DP2 method only (default = 7)

    :rtype: vector
    :return gam: describing the warping function used to align q2 with q1

    """

    if penalty == "l2gam" and (method == "DP" or method == "DP2"):
        raise Exception("penalty not implemented")
    if penalty == "l2psi" and (method == "DP" or method == "DP2"):
        raise Exception("penalty not implemented")
    if penalty == "geodesic" and (method == "DP" or method == "DP2"):
        raise Exception("penalty not implemented")

    if method == "DP":
        if q1.ndim == 1 and q2.ndim == 1:
            gam = orN.coptimum_reparam(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam
            )

        if q1.ndim == 1 and q2.ndim == 2:
            gam = orN.coptimum_reparam_N(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam
            )

        if q1.ndim == 2 and q2.ndim == 2:
            gam = orN.coptimum_reparam_N2(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam
            )
    elif method == "DP2":
        if q1.ndim == 1 and q2.ndim == 1:
            gam = orN2.coptimum_reparam(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam, grid_dim
            )

        if q1.ndim == 1 and q2.ndim == 2:
            gam = orN2.coptimum_reparamN(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam, grid_dim
            )

        if q1.ndim == 2 and q2.ndim == 2:
            gam = orN2.coptimum_reparamN2(
                ascontiguousarray(q1), time, ascontiguousarray(q2), lam, grid_dim
            )
    elif method == "RBFGS":
        if q1.ndim == 1 and q2.ndim == 1:
            time = linspace(0, 1, q1.shape[0])
            obj = rlbfgs(q1, q2, time)
            obj.solve(lam=lam, penalty=penalty)
            gam = obj.gammaOpt

        if q1.ndim == 1 and q2.ndim == 2:
            gam = zeros(q2.shape)
            time = linspace(0, 1, q1.shape[0])
            for i in range(0, q2.shape[1]):
                obj = rlbfgs(q1, q2[:, i], time)
                obj.solve(lam=lam, penalty=penalty)
                gam[:, i] = obj.gammaOpt

        if q1.ndim == 2 and q2.ndim == 2:
            gam = zeros(q2.shape)
            time = linspace(0, 1, q1.shape[0])
            for i in range(0, q2.shape[1]):
                obj = rlbfgs(q1[:, i], q2[:, i], time)
                obj.solve(lam=lam, penalty=penalty)
                gam[:, i] = obj.gammaOpt
    elif method == "cRBFGS":
        import crbfgs as cr
        if penalty == "roughness":
            pen = 0
        elif penalty == "l2gam":
            pen = 1
        elif penalty == "l2psi":
            pen = 2
        elif penalty == "geodesic":
            pen = 3
        else:
            raise Exception("penalty not implemented")

        if q1.ndim == 1 and q2.ndim == 1:
            time = linspace(0, 1, q1.shape[0])
            gam = cr.rlbfgs(
                ascontiguousarray(q1),
                ascontiguousarray(q2),
                ascontiguousarray(time),
                30,
                lam,
                pen,
            )

        if q1.ndim == 1 and q2.ndim == 2:
            gam = zeros(q2.shape)
            time = linspace(0, 1, q1.shape[0])
            for i in range(0, q2.shape[1]):
                gam[:, i] = cr.rlbfgs(
                    ascontiguousarray(q1),
                    ascontiguousarray(q2[:, i]),
                    ascontiguousarray(time),
                    30,
                    lam,
                    pen,
                )

        if q1.ndim == 2 and q2.ndim == 2:
            gam = zeros(q2.shape)
            time = linspace(0, 1, q1.shape[0])
            for i in range(0, q2.shape[1]):
                gam[:, i] = cr.rlbfgs(
                    ascontiguousarray(q1[:, i]),
                    ascontiguousarray(q2[:, i]),
                    ascontiguousarray(time),
                    30,
                    lam,
                    pen,
                )

    else:
        raise Exception("Invalid Optimization Method")

    gam = (gam - gam.min()) / (gam.max() - gam.min())
    return gam


def optimum_reparam_pair(q, time, q1, q2, lam=0.0):
    """
    calculates the warping to align srsf pair q1 and q2 to q

    :param q: vector of size N or array of NxM samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q1: vector of size N or array of NxM samples samples of second SRSF
    :param q2: vector of size N or array of NxM samples samples of second SRSF
    :param lam: controls the amount of elasticity (default = 0.0)

    :rtype: vector
    :return gam: describing the warping function used to align q2 with q1

    """
    if q1.ndim == 1 and q2.ndim == 1:
        q_c = column_stack((q1, q2))
        gam = orN.coptimum_reparam_pair(
            ascontiguousarray(q), time, ascontiguousarray(q_c), lam
        )

    if q1.ndim == 2 and q2.ndim == 2:
        gam = orN.coptimum_reparamN2_pair(
            ascontiguousarray(q),
            time,
            ascontiguousarray(q1),
            ascontiguousarray(q2),
            lam,
        )

    return gam


def distmat(f, f1, time, idx, method):
    N = f.shape[1]
    dp = zeros(N)
    da = zeros(N)
    for jj in range(N):
        Dy, Dx = elastic_distance(f[:, jj], f1, time, method)

        da[jj] = Dy
        dp[jj] = Dx

    return (da, dp)


def elastic_depth(f, time, method="DP2", lam=0.0, parallel=True):
    """
    calculates the elastic depth between functions in matrix f

    :param f: matrix of size MxN (M time points for N functions)
    :param time: vector of size M describing the sample points
    :param method: method to apply optimization (default="DP2")
                   options are "DP","DP2","RBFGS","cRBFGS"
    :param lam: controls the elasticity (default = 0.0)

    :rtype: scalar
    :return amp: amplitude depth
    :return phase: phase depth

    """

    obs, fns = f.shape

    amp_dist = zeros((fns, fns))
    phs_dist = zeros((fns, fns))

    if parallel:
        out = Parallel(n_jobs=-1)(
            delayed(distmat)(f, f[:, n], time, n, method) for n in range(fns)
        )
        for i in range(0, fns):
            amp_dist[i, :] = out[i][0]
            phs_dist[i, :] = out[i][1]
    else:
        for i in range(0, fns):
            amp_dist[i, :], phs_dist[i, :] = distmat(f, f[:, i], time, i, method)

    amp_dist = amp_dist + amp_dist.T
    phs_dist = phs_dist + phs_dist.T

    amp = 1 / (1 + median(amp_dist, axis=0))
    phase = 1 / (1 + median(phs_dist, axis=0))
    phase = ((2 + pi) / pi) * (phase - 2 / (2 + pi))

    return amp, phase


def elastic_distance(
    f1, f2, time, method="DP2", lam=0.0, alpha=None, return_dt_only=True
):
    """ "
    calculates the distances between function, where f1 is aligned to
    f2. In other words calculates the elastic distances

    :param f1: vector of size N
    :param f2: vector of size N
    :param time: vector of size N describing the sample points
    :param method: method to apply optimization (default="DP2")
                   options are "DP","DP2","RBFGS","cRBFGS"
    :param lam: controls the elasticity (default = 0.0)
    :param alpha: makes alpha * dx + (1-alpha) * dy
    :param return_dt_only: returns only dt if alpha is set

    :rtype: scalar
    :return Dy: amplitude distance
    :return Dx: phase distance
    :return Dt: combined distance

    """
    q1 = f_to_srsf(f1, time)
    q2 = f_to_srsf(f2, time)

    gam = optimum_reparam(q1, time, q2, method, lam)
    qw = warp_q_gamma(time, q2, gam)

    Dy = sqrt(trapz((qw - q1) ** 2, time))
    M = time.shape[0]

    time1 = linspace(0, 1, M)
    binsize = mean(diff(time1))
    psi = sqrt(gradient(gam, binsize))
    q1dotq2 = trapz(psi, time1)
    if q1dotq2 > 1:
        q1dotq2 = 1
    elif q1dotq2 < -1:
        q1dotq2 = -1

    Dx = real(arccos(q1dotq2))

    if alpha is not None:
        Dt = alpha * Dx + (1 - alpha) * Dy
        if return_dt_only:
            return Dt
        else:
            return Dy, Dx, Dt
    else:
        return Dy, Dx


def invertGamma(gam):
    """
    finds the inverse of the diffeomorphism gamma

    :param gam: vector describing the warping function

    :rtype: vector
    :return gamI: inverse of gam

    """
    N = gam.size
    x = linspace(0, 1, N)
    s = interp1d(gam, x)
    gamI = s(x)
    gamI = (gamI - gamI[0]) / (gamI[-1] - gamI[0])
    return gamI


def SqrtMeanInverse(gam):
    """
    finds the inverse of the mean of the set of the diffeomorphisms gamma

    :param gam: numpy ndarray of shape (M,N) of N warping functions
                with M samples

    :rtype: vector
    :return gamI: inverse of gam


    """
    (T, n) = gam.shape
    time = linspace(0, 1, T)
    binsize = mean(diff(time))
    psi = zeros((T, n))
    for k in range(0, n):
        psi[:, k] = sqrt(gradient(gam[:, k], binsize))

    # Find Direction
    mnpsi = psi.mean(axis=1)
    a = mnpsi.repeat(n)
    d1 = a.reshape(T, n)
    d = (psi - d1) ** 2
    dqq = sqrt(d.sum(axis=0))
    min_ind = dqq.argmin()
    mu = psi[:, min_ind]
    maxiter = 501
    tt = 1
    lvm = zeros(maxiter)
    vec = zeros((T, n))
    stp = 0.3
    itr = 0

    for i in range(0, n):
        out, theta = geo.inv_exp_map(mu, psi[:, i])
        vec[:, i] = out

    vbar = vec.mean(axis=1)
    lvm[itr] = geo.L2norm(vbar)

    while (lvm[itr] > 0.00000001) and (itr < maxiter):
        mu = geo.exp_map(mu, stp * vbar)
        itr += 1
        for i in range(0, n):
            out, theta = geo.inv_exp_map(mu, psi[:, i])
            vec[:, i] = out

        vbar = vec.mean(axis=1)
        lvm[itr] = geo.L2norm(vbar)

    gam_mu = cumtrapz(mu * mu, time, initial=0)
    gam_mu = (gam_mu - gam_mu.min()) / (gam_mu.max() - gam_mu.min())
    gamI = invertGamma(gam_mu)
    return gamI


def SqrtMean(gam, parallel=False, cores=-1):
    """
    calculates the srsf of warping functions with corresponding shooting
    vectors

    :param gam: numpy ndarray of shape (M,N) of M warping functions
                with N samples
    :param parallel: run in parallel (default = F)
    :param cores: number of cores for parallel (default = -1 (all))

    :rtype: 2 numpy ndarray and vector
    :return mu: Karcher mean psi function
    :return gam_mu: vector of dim N which is the Karcher mean warping
                    function
    :return psi: numpy ndarray of shape (M,N) of M SRSF of the warping
                 functions
    :return vec: numpy ndarray of shape (M,N) of M shooting vectors

    """

    (T, n) = gam.shape
    time = linspace(0, 1, T)
    binsize = mean(diff(time))
    psi = zeros((T, n))
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(gradient)(gam[:, k], binsize) for k in range(n)
        )
        psi = array(out)
        psi = psi.transpose()
        psi = sqrt(psi)
    else:
        for k in range(0, n):
            psi[:, k] = sqrt(gradient(gam[:, k], binsize))

    # Find Direction
    mnpsi = psi.mean(axis=1)
    a = mnpsi.repeat(n)
    d1 = a.reshape(T, n)
    d = (psi - d1) ** 2
    dqq = sqrt(d.sum(axis=0))
    min_ind = dqq.argmin()
    mu = psi[:, min_ind]
    maxiter = 501
    lvm = zeros(maxiter)
    vec = zeros((T, n))
    stp = 0.3
    itr = 0

    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(inv_exp_map_sub)(mu, psi[:, i]) for i in range(n)
        )
        vec = array(out)
        vec = vec.transpose()
    else:
        for i in range(0, n):
            out, theta = geo.inv_exp_map(mu, psi[:, i])
            vec[:, i] = out

    vbar = vec.mean(axis=1)
    lvm[itr] = geo.L2norm(vbar)

    while (lvm[itr] > 0.00000001) and (itr < maxiter):
        mu = geo.exp_map(mu, stp * vbar)
        itr += 1
        if parallel:
            out = Parallel(n_jobs=cores)(
                delayed(inv_exp_map_sub)(mu, psi[:, i]) for i in range(n)
            )
            vec = array(out)
            vec = vec.transpose()
        else:
            for i in range(0, n):
                out, theta = geo.inv_exp_map(mu, psi[:, i])
                vec[:, i] = out

        vbar = vec.mean(axis=1)
        lvm[itr] = geo.L2norm(vbar)

    gam_mu = cumtrapz(mu * mu, time, initial=0)
    gam_mu = (gam_mu - gam_mu.min()) / (gam_mu.max() - gam_mu.min())

    return mu, gam_mu, psi, vec


def inv_exp_map_sub(mu, psi):
    out, theta = geo.inv_exp_map(mu, psi)
    return out


def SqrtMedian(gam):
    """
    calculates the median srsf of warping functions with corresponding
    shooting vectors

    :param gam: numpy ndarray of shape (M,N) of M warping functions
                with N samples

    :rtype: 2 numpy ndarray and vector
    :return gam_median: Karcher median warping function
    :return psi_meidan: vector of dim N which is the Karcher median srsf
                        function
    :return psi: numpy ndarray of shape (M,N) of M SRSF of the warping
                 functions
    :return vec: numpy ndarray of shape (M,N) of M shooting vectors

    """

    (T, n) = gam.shape
    time = linspace(0, 1, T)

    # Initialization
    psi_median = ones(T)
    r = 1
    stp = 0.3
    maxiter = 501
    vbar_norm = zeros(maxiter + 1)

    # compute psi function
    binsize = mean(diff(time))
    psi = zeros((T, n))
    v = zeros((T, n))
    vtil = zeros((T, n))
    d = zeros(n)
    dtil = zeros(n)
    for k in range(0, n):
        psi[:, k] = sqrt(gradient(gam[:, k], binsize))
        v[:, k], d[k] = geo.inv_exp_map(psi_median, psi[:, k])
        vtil[:, k] = v[:, k] / d[k]
        dtil[k] = 1 / d[k]

    vbar = vtil.sum(axis=1) * dtil.sum() ** (-1)
    vbar_norm[r] = geo.L2norm(vbar)

    # compute phase median by iterative algorithm
    while (vbar_norm[r] > 0.00000001) and (r < maxiter):
        psi_median = geo.exp_map(psi_median, stp * vbar)
        r += 1
        for k in range(0, n):
            v[:, k], tmp = geo.inv_exp_map(psi_median, psi[:, k])
            d[k] = arccos(geo.inner_product(psi_median, psi[:, k]))
            vtil[:, k] = v[:, k] / d[k]
            dtil[k] = 1 / d[k]

        vbar = vtil.sum(axis=1) * dtil.sum() ** (-1)
        vbar_norm[r] = geo.L2norm(vbar)

    vec = v
    gam_median = cumtrapz(psi_median**2, time, initial=0.0)
    gam_median = (gam_median - gam_median.min()) / (gam_median.max() - gam_median.min())

    return gam_median, psi_median, psi, vec


def cumtrapzmid(x, y, c, mid):
    """
    cumulative trapezoidal numerical integration taken from midpoint

    :param x: vector of size N describing the time samples
    :param y: vector of size N describing the function
    :param c: midpointtic
    :param mid: midpiont location

    :rtype: vector
    :return fa: cumulative integration

    """
    a = x.shape[0]

    # case < mid
    fa = zeros(a)
    tmpx = x[0:mid]
    tmpy = y[0:mid]
    tmp = c + cumtrapz(tmpy[::-1], tmpx[::-1], initial=0)
    fa[0:mid] = tmp[::-1]

    # case >= mid
    fa[mid:a] = c + cumtrapz(y[mid - 1 : a - 1], x[mid - 1 : a - 1], initial=0)

    return fa


def rgam(N, sigma, num, mu_gam=None):
    """
    Generates random warping functions

    :param N: length of warping function
    :param sigma: variance of warping functions
    :param num: number of warping functions
    :param mu_gam mean warping function (default identity)
    :return: gam: numpy ndarray of warping functions

    """
    gam = zeros((N, num))

    time = linspace(0, 1, N)
    binsize = diff(time)
    binsize = binsize.mean()
    if mu_gam is None:
        mu = sqrt(gradient(time, binsize))
    else:
        mu = sqrt(gradient(mu_gam, binsize))

    omega = 2 * pi
    for k in range(0, num):
        alpha_i = rn.normal(scale=sigma)
        v = alpha_i * ones(N)
        cnt = 1
        for l in range(2, 4):
            alpha_i = rn.normal(scale=sigma)

            # odd
            if l % 2 != 0:
                v = v + alpha_i * sqrt(2) * cos(cnt * omega * time)
                cnt += 1

            # even
            if l % 2 == 0:
                v = v + alpha_i * sqrt(2) * sin(cnt * omega * time)

        psi = geo.exp_map(mu.ravel(), v.ravel())
        gam0 = cumtrapz(psi * psi, time, initial=0)
        gam[:, k] = (gam0 - gam0.min()) / (gam0.max() - gam0.min())

    return gam


def outlier_detection(q, time, mq, k=1.5):
    """
    calculates outlier's using geodesic distances of the SRSFs from the median

    :param q: numpy ndarray of N x M of M SRS functions with N samples
    :param time: vector of size N describing the sample points
    :param mq: median calculated using :func:`time_warping.srsf_align`
    :param k: cutoff threshold (default = 1.5)

    :return: q_outlier: outlier functions

    """
    N = q.shape[1]
    ds = zeros(N)
    for kk in range(0, N):
        ds[kk] = sqrt(trapz((mq - q[:, kk]) ** 2, time))

    quartile_range = mquantiles(ds)
    IQR = quartile_range[2] - quartile_range[0]

    thresh = quartile_range[2] + k * IQR

    ind = (ds > thresh).nonzero()

    q_outlier = q[:, ind]

    return q_outlier


def randomGamma(gam, num):
    """
    generates random warping functions

    :param gam: numpy ndarray of N x M of M of warping functions
    :param num: number of random functions

    :return: rgam: random warping functions

    """
    mu, gam_mu, psi, vec = SqrtMean(gam)
    K = cov(vec)

    U, s, V = svd(K)
    n = 5
    TT = vec.shape[0] + 1
    vm = vec.mean(axis=1)

    rgam = zeros((TT, num))
    for k in range(0, num):
        a = rn.standard_normal(n)
        v = zeros(vm.size)
        for i in range(0, n):
            v = v + a[i] * sqrt(s[i]) * U[:, i]

        vn = norm(v) / sqrt(TT)
        psi = cos(vn) * mu + sin(vn) * v / vn
        tmp = zeros(TT)
        tmp[1:TT] = cumsum(psi * psi) / TT
        rgam[:, k] = (tmp - tmp[0]) / (tmp[-1] - tmp[0])

    return rgam


def update_progress(progress):
    """
    This function creates a progress bar

    :param progress: fraction of progress

    """
    barLength = 20  # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    text = "\rPercent: [{0}] {1}% {2}".format(
        "#" * block + "-" * (barLength - block), progress * 100, status
    )
    sys.stdout.write(text)
    sys.stdout.flush()


def diffop(n, binsize=1):
    """
    Creates a second order differential operator

    :param n: dimension
    :param binsize: dx (default = 1)

    :rtype: numpy ndarray
    :return m: matrix describing differential operator

    """
    m = diagflat(ones(n - 1), k=1) + diagflat(ones(n - 1), k=-1) + diagflat(2 * ones(n))
    m = inner(m.transpose(), m)
    m[0, 0] = 6
    m[-1, -1] = 6
    m /= binsize**4.0

    return m


def geigen(Amat, Bmat, Cmat):
    """
    generalized eigenvalue problem of the form

    max tr L'AM / sqrt(tr L'BL tr M'CM) w.r.t. L and M

    :param Amat numpy ndarray of shape (M,N)
    :param Bmat numpy ndarray of shape (M,N)
    :param Bmat numpy ndarray of shape (M,N)

    :rtype: numpy ndarray
    :return values: eigenvalues
    :return Lmat: left eigenvectors
    :return Mmat: right eigenvectors

    """
    if Bmat.shape[0] != Bmat.shape[1]:
        print("BMAT is not square.\n")
        sys.exit(1)

    if Cmat.shape[0] != Cmat.shape[1]:
        print("CMAT is not square.\n")
        sys.exit(1)

    p = Bmat.shape[0]
    q = Cmat.shape[0]

    tmp = fabs(Bmat - Bmat.transpose())
    tmp1 = fabs(Bmat)
    if tmp.max() / tmp1.max() > 1e-10:
        print("BMAT not symmetric..\n")
        sys.exit(1)

    tmp = fabs(Cmat - Cmat.transpose())
    tmp1 = fabs(Cmat)
    if tmp.max() / tmp1.max() > 1e-10:
        print("CMAT not symmetric..\n")
        sys.exit(1)

    Bmat = (Bmat + Bmat.transpose()) / 2.0
    Cmat = (Cmat + Cmat.transpose()) / 2.0
    Bfac = cholesky(Bmat)
    Cfac = cholesky(Cmat)
    Bfacinv = inv(Bfac)
    Bfacinvt = Bfacinv.transpose()
    Cfacinv = inv(Cfac)
    Dmat = Bfacinvt.dot(Amat).dot(Cfacinv)
    if p >= q:
        u, d, vh = svd(Dmat)
        values = d
        Lmat = Bfacinv.dot(u)
        Mmat = Cfacinv.dot(vh)
    else:
        u, d, vh = svd(Dmat.transpose())
        values = d
        Lmat = Bfacinv.dot(u)
        Mmat = Cfacinv.dot(vh)

    return values, Lmat, Mmat


def innerprod_q(time, q1, q2):
    """
    calculates the innerproduct between two srsfs

    :param time vector descrbing time samples
    :param q1 vector of srsf 1
    :param q2 vector of srsf 2

    :rtype: scalar
    :return val: inner product value

    """
    val = trapz(q1 * q2, time)
    return val


def warp_q_gamma(time, q, gam):
    """
    warps a srsf q by gam

    :param time vector describing time samples
    :param q vector describing srsf
    :param gam vector describing warping function

    :rtype: numpy ndarray
    :return q_temp: warped srsf

    """
    M = gam.size
    gam_dev = gradient(gam, 1 / double(M - 1))
    tmp = interp((time[-1] - time[0]) * gam + time[0], time, q)

    q_temp = tmp * sqrt(gam_dev)

    return q_temp


def warp_f_gamma(time, f, gam):
    """
    warps a function f by gam

    :param time vector describing time samples
    :param q vector describing srsf
    :param gam vector describing warping function

    :rtype: numpy ndarray
    :return f_temp: warped srsf

    """
    f_temp = interp((time[-1] - time[0]) * gam + time[0], time, f)

    return f_temp


def f_K_fold(Nobs, K=5):
    """
    generates sample indices for K-fold cross validation

    :param Nobs number of observations
    :param K number of folds

    :rtype: numpy ndarray
    :return train: train indexes (Nobs*(K-1)/K X K)
    :return test: test indexes (Nobs*(1/K) X K)

    """
    rs = rn.uniform(size=Nobs)
    ids = rs.ravel().argsort()
    k = Nobs * arange(1, K) / K
    tmp = append(k.repeat(2), Nobs)
    tmp = insert(tmp, 0, 0)
    k = tmp.reshape((K, 2))
    k[:, 0] = k[:, 0] + 1
    k = k - 1
    train = zeros((Nobs * (K - 1) / K, K))
    test = zeros((Nobs * 1 / K, K))
    for ii in range(0, K):
        tf = vectorize(lambda x: x in arange(k[ii, 0], k[ii, 1] + 1))(arange(0, Nobs))
        train[:, ii] = ids[not tf]
        test[:, ii] = ids[tf]

    return train, test


def zero_crossing(Y, q, bt, time, y_max, y_min, gmax, gmin):
    """
    finds zero-crossing of optimal gamma, gam = s*gmax + (1-s)*gmin
    from elastic regression model

    :param Y: response
    :param q: predicitve function
    :param bt: basis function
    :param time: time samples
    :param y_max: maximum repsonse for warping function gmax
    :param y_min: minimum response for warping function gmin
    :param gmax: max warping function
    :param gmin: min warping fucntion

    :rtype: numpy array
    :return gamma: optimal warping function

    """
    # simple iterative method based on intermediate theorem
    max_itr = 100
    a = zeros(max_itr)
    a[0] = 1
    f = zeros(max_itr)
    f[0] = y_max - Y
    f[1] = y_min - Y
    mrp = f[0]
    mrn = f[1]
    mrp_ind = 0  # most recent positive index
    mrn_ind = 1  # most recent negative index

    for ii in range(2, max_itr):
        x1 = a[mrp_ind]
        x2 = a[mrn_ind]
        y1 = mrp
        y2 = mrn
        a[ii] = (x1 * y2 - x2 * y1) / float(y2 - y1)

        gam_m = a[ii] * gmax + (1 - a[ii]) * gmin
        qtmp = warp_q_gamma(time, q, gam_m)
        f[ii] = trapz(qtmp * bt, time) - Y

        if fabs(f[ii]) < 1e-5:
            break
        elif f[ii] > 0:
            mrp = f[ii]
            mrp_ind = ii
        else:
            mrn = f[ii]
            mrn_ind = ii

    gamma = a[ii] * gmax + (1 - a[ii]) * gmin

    return gamma


def resamplefunction(x, n):
    """
    resample function using n points

    :param x: functions
    :param n: number of points

    :rtype: numpy array
    :return xn: resampled function

    """
    T = x.shape[0]
    xn = interp(arange(0, n) / double(n - 1), arange(0, T) / double(T - 1), x)
    return xn


def basis_fourierd(f_domain, numBasis):
    result = zeros((f_domain.shape[0], 2 * numBasis))
    for i in range(0, 2 * numBasis):
        j = ceil(i / 2)
        if mod(i, 2) == 1:
            result[:, i] = 1 / sqrt(pi) * sin(2 * j * pi * f_domain)

        if mod(i, 2) == 0:
            result[:, i] = 1 / sqrt(pi) * cos(2 * j * pi * f_domain)

    out = {"x": f_domain, "matrix": result}

    return out


def basis_fourier(f_domain, numBasis, fourier_p):
    result = zeros((f_domain.shape[0], 2 * numBasis))
    for i in range(0, 2 * numBasis):
        j = ceil(i / 2)
        if mod(i, 2) == 1:
            result[:, i] = sqrt(2) * sin(2 * j * pi * f_domain / fourier_p)

        if mod(i, 2) == 0:
            result[:, i] = sqrt(2) * cos(2 * j * pi * f_domain / fourier_p)

    out = {"x": f_domain, "matrix": result}

    return out


def legendre(N, X):
    matrixReturn = zeros((N + 1, X.shape[0]))
    for i in enumerate(X):
        currValues = sp.lpmn(N, N, i[1])
        matrixReturn[:, i[0]] = array([j[N] for j in currValues[0]])
    return matrixReturn


def basis_legendre(f_domain, numBasis, fourier_p):
    result = zeros((f_domain.shape[0], 2 * numBasis))
    for i in range(0, 2 * numBasis):
        f_domain_scaled = 2 * (f_domain / fourier_p) - 1
        tmp = legendre(i + 1, f_domain_scaled)
        result[:, i] = tmp[0, :]

    out = {"x": f_domain, "matrix": result}

    return out


def exp2corr(sigma2, phi, ds):
    out = sigma2 * exp(-(ds**2) / (2 * phi**2))
    return out


def exp2corr2(phi, ds):
    out = exp(-(ds**2) / (2 * phi**2))
    return out


def statsFun(vec):
    a = percentile(vec, 100 * 0.025, axis=1)
    b = percentile(vec, 100 * 0.975, axis=1)
    out = column_stack((a, b))
    return out


def f_exp1(g):
    out = bay.bcalcY(f_L2norm(g), g)
    return out


def f_L2norm(f):
    x = linspace(0, 1, f.shape[0])
    out = bay.border_l2norm(x, f)
    return out


def f_basistofunction(f_domain, coefconst, coef, basis):
    if basis["matrix"].shape[1] < coef.shape[0]:
        raise Exception("In f_basistofunction, #coefficients exceeds # basis functions")

    result = dot(basis["matrix"], coef) + coefconst
    result = f_predictfunction(result, f_domain, 0)
    return result


def f_predictfunction(f, at, deriv):
    x = linspace(0, 1, f.shape[0])
    if deriv == 0:
        interp = interp1d(x, f, bounds_error=False, fill_value="extrapolate")
        result = interp(at)

    if deriv == 1:
        iterp = interp1d(x, f, bounds_error=False, fill_value="extrapolate")
        fmod = iterp(at)
        diffy1 = hstack((0, diff(fmod)))
        diffy2 = hstack((diff(fmod), 0))
        diffx1 = hstack((0, diff(at)))
        diffx2 = hstack((diff(at), 0))

        result = (diffy2 + diffy1) / (diffx2 + diffx1)

    return result


def f_psimean(x, y):
    rmy = y.mean(axis=1)
    result = rmy / f_L2norm(rmy)
    return result


def f_phiinv(psi):
    f_domain = linspace(0, 1, psi.shape[0])
    result = insert(bay.bcuL2norm2(f_domain, psi), 0, 0)
    return result


def norm_gam(gam):
    gam = (gam - gam[0]) / (gam[-1] - gam[0])
    return gam


def mrdivide(a, b):
    c = dot(a, pinv(b))
    return c


def rlbfgs_dist(q1, q2):
    import crbfgs as cr
    q1 = q1.copy(order="C")
    q2 = q2.copy(order="C")
    d = cr.rlbfgs_dist(q1, q2)
    return d



"""
Vertical and Horizontal Functional Principal Component Analysis using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.geometry as geo
from scipy.linalg import norm, svd
from scipy.integrate import trapz, cumtrapz
from scipy.optimize import fminbound
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
import fdasrsf.plot_style as plot
import collections


class fdavpca:
    """
    This class provides vertical fPCA using the
    SRVF framework

    Usage:  obj = fdavpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param q_pca: srvf principal directions
    :param f_pca: f principal directions
    :param latent: latent values
    :param coef: principal coefficients
    :param id: point used for f(0)
    :param mqn: mean srvf
    :param U: eigenvectors
    :param stds: geodesic directions
    :param new_coef: principal coefficients of new data

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp

    def calc_fpca(self, no=3, var_exp=None, id=None, stds=np.arange(-1, 2)):
        """
        This function calculates vertical functional principal component
        analysis on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param id: point to use for f(0) (default = midpoint)
        :param stds: number of standard deviations along geodesic to compute
                     (default = -1,0,1)
        :type no: int
        :type id: int

        :rtype: fdavpca object containing
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        fn = self.warp_data.fn
        time = self.warp_data.time
        qn = self.warp_data.qn

        M = time.shape[0]
        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = M

        if id is None:
            mididx = int(np.round(M / 2))
        else:
            mididx = id

        Nstd = stds.shape[0]

        # FPCA
        mq_new = qn.mean(axis=1)
        N = mq_new.shape[0]
        m_new = np.sign(fn[mididx, :]) * np.sqrt(np.abs(fn[mididx, :]))
        mqn = np.append(mq_new, m_new.mean())
        self.mqn2 = mqn
        qn2 = np.vstack((qn, m_new))
        K = np.cov(qn2)

        U, s, V = svd(K)
        stdS = np.sqrt(s)

        # compute the PCA in the q domain
        q_pca = np.ndarray(shape=(N + 1, Nstd, no), dtype=float)
        for k in range(0, no):
            for l in range(0, Nstd):
                q_pca[:, l, k] = mqn + stds[l] * stdS[k] * U[:, k]

        # compute the correspondence in the f domain
        f_pca = np.ndarray(shape=(N, Nstd, no), dtype=float)
        for k in range(0, no):
            for l in range(0, Nstd):
                f_pca[:, l, k] = uf.cumtrapzmid(
                    time,
                    q_pca[0:N, l, k] * np.abs(q_pca[0:N, l, k]),
                    np.sign(q_pca[N, l, k]) * (q_pca[N, l, k] ** 2),
                    mididx,
                )
            fbar = fn.mean(axis=1)
            fsbar = f_pca[:, :, k].mean(axis=1)
            err = np.transpose(np.tile(fbar - fsbar, (Nstd, 1)))
            f_pca[:, :, k] += err

        N2 = qn.shape[1]
        c = np.zeros((N2, no))
        for k in range(0, no):
            for l in range(0, N2):
                c[l, k] = sum((np.append(qn[:, l], m_new[l]) - mqn) * U[:, k])

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / sum(s)
            no = np.argwhere(cumm_coef <= var_exp)[-1]

        self.q_pca = q_pca
        self.f_pca = f_pca
        self.latent = s[0:no]
        self.coef = c[:, 0:no]
        self.U = U[:, 0:no]
        self.id = mididx
        self.mqn = mqn
        self.time = time
        self.stds = stds
        self.no = no

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        fn = np.zeros((M, n))
        qn = np.zeros((M, n))
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])
            fn[:, ii] = fs.warp_f_gamma(self.time, f[:, ii], gam[:, ii])
            qn[:, ii] = fs.f_to_srsf(fn[:, ii], self.time)

        U = self.U
        no = U.shape[1]

        m_new = np.sign(fn[self.id, :]) * np.sqrt(np.abs(fn[self.id, :]))
        qn1 = np.vstack((qn, m_new))

        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                a[i, j] = sum((qn1[:, i] - self.mqn2) * U[:, j])

        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic vertical fPCA result
        Usage: obj.plot()
        """

        no = self.no
        Nstd = self.stds.shape[0]
        N = self.time.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        cl = sorted(CBcdict.keys())
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(2, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[0, k1]
                if k > (no - 1):
                    break

                for l in range(0, Nstd):
                    axt.plot(self.time, self.q_pca[0:N, l, k], color=CBcdict[cl[l]])

                axt.set_title("q domain: PD %d" % (k + 1))

                axt = ax[1, k1]
                for l in range(0, Nstd):
                    axt.plot(self.time, self.f_pca[:, l, k], color=CBcdict[cl[l]])

                axt.set_title("f domain: PD %d" % (k + 1))

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent) / sum(self.latent)
        N = self.latent.shape[0]
        idx = np.arange(0, N) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


class fdahpca:
    """
    This class provides horizontal fPCA using the
    SRVF framework

    Usage:  obj = fdahpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param gam_pca: warping functions principal directions
    :param psi_pca: srvf principal directions
    :param latent: latent values
    :param U: eigenvectors
    :param coef: coefficients
    :param vec: shooting vectors
    :param mu: Karcher Mean
    :param tau: principal directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp

    def calc_fpca(self, no=3, var_exp=None, stds=np.arange(-1, 2)):
        """
        This function calculates horizontal functional principal component
        analysis on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param stds: number of standard deviations along geodesic to compute
                     (default = -1,0,1)
        :type no: int

        :rtype: fdahpca object of numpy ndarray
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        # Calculate Shooting Vectors
        gam = self.warp_data.gam
        mu, gam_mu, psi, vec = uf.SqrtMean(gam)
        TT = self.warp_data.time.shape[0]

        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = TT

        # TFPCA
        K = np.cov(vec)

        U, s, V = svd(K)
        vm = vec.mean(axis=1)
        self.vm = vm
        self.mu_psi = mu

        gam_pca = np.ndarray(shape=(stds.shape[0], mu.shape[0], no), dtype=float)
        psi_pca = np.ndarray(shape=(stds.shape[0], mu.shape[0], no), dtype=float)
        for j in range(0, no):
            cnt = 0
            for k in stds:
                v = k * np.sqrt(s[j]) * U[:, j]
                vn = norm(v) / np.sqrt(TT)
                if vn < 0.0001:
                    psi_pca[cnt, :, j] = mu
                else:
                    psi_pca[cnt, :, j] = np.cos(vn) * mu + np.sin(vn) * v / vn

                tmp = cumtrapz(
                    psi_pca[cnt, :, j] * psi_pca[cnt, :, j],
                    np.linspace(0, 1, TT),
                    initial=0,
                )
                gam_pca[cnt, :, j] = (tmp - tmp[0]) / (tmp[-1] - tmp[0])
                cnt += 1

        N2 = gam.shape[1]
        c = np.zeros((N2, no))
        for k in range(0, no):
            for i in range(0, N2):
                c[i, k] = np.dot(vec[:, i] - vm, U[:, k])

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / sum(s)
            no = np.argwhere(cumm_coef <= var_exp)[-1]

        self.gam_pca = gam_pca
        self.psi_pca = psi_pca
        self.U = U[:, 0:no]
        self.coef = c[:, 0:no]
        self.latent = s[0:no]
        self.gam_mu = gam_mu
        self.psi_mu = mu
        self.vec = vec
        self.no = no
        self.stds = stds

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])

        U = self.U
        no = U.shape[1]

        mu_psi = self.mu_psi
        vec = np.zeros((M, n))
        psi = np.zeros((M, n))
        binsize = np.mean(np.diff(self.time))
        for i in range(0, n):
            psi[:, i] = np.sqrt(np.gradient(gam[:, i], binsize))
            out, theta = fs.inv_exp_map(mu_psi, psi[:, i])
            vec[:, i] = out

        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                a[i, j] = np.dot(vec[:, i] - self.vm, U[:, j])

        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic horizontal fPCA results

        Usage: obj.plot()
        """

        no = self.no
        TT = self.warp_data.time.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(1, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[k1]
                if k > (no - 1):
                    break

                tmp = self.gam_pca[:, :, k]
                axt.plot(np.linspace(0, 1, TT), tmp.transpose())
                axt.set_title("PD %d" % (k + 1))
                axt.set_aspect("equal")

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent[0:no]) / sum(self.latent[0:no])
        idx = np.arange(0, no) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


class fdajpca:
    """
    This class provides joint fPCA using the
    SRVF framework

    Usage:  obj = fdajpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param q_pca: srvf principal directions
    :param f_pca: f principal directions
    :param latent: latent values
    :param coef: principal coefficients
    :param id: point used for f(0)
    :param mqn: mean srvf
    :param U: eigenvectors
    :param mu_psi: mean psi
    :param mu_g: mean g
    :param C: scaling value
    :param stds: geodesic directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp
        self.M = fdawarp.time.shape[0]

    def calc_fpca(
        self,
        no=3,
        var_exp=None,
        stds=np.arange(-1.0, 2.0),
        id=None,
        parallel=False,
        cores=-1,
    ):
        """
        This function calculates joint functional principal component analysis
        on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param id: point to use for f(0) (default = midpoint)
        :param stds: number of standard deviations along gedoesic to compute
                     (default = -1,0,1)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :type no: int
        :type id: int
        :type parallel: bool
        :type cores: int

        :rtype: fdajpca object of numpy ndarray
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        fn = self.warp_data.fn
        time = self.warp_data.time
        qn = self.warp_data.qn
        q0 = self.warp_data.q0
        gam = self.warp_data.gam

        M = time.shape[0]
        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = M

        if id is None:
            mididx = int(np.round(M / 2))
        else:
            mididx = id

        Nstd = stds.shape[0]

        # set up for fPCA in q-space
        mq_new = qn.mean(axis=1)
        m_new = np.sign(fn[mididx, :]) * np.sqrt(np.abs(fn[mididx, :]))
        mqn = np.append(mq_new, m_new.mean())
        qn2 = np.vstack((qn, m_new))

        # calculate vector space of warping functions
        mu_psi, gam_mu, psi, vec = uf.SqrtMean(gam, parallel, cores)

        # joint fPCA
        C = fminbound(find_C, 0, 1e4, (qn2, vec, q0, no, mu_psi, parallel, cores))
        qhat, gamhat, a, U, s, mu_g, g, cov = jointfPCAd(
            qn2, vec, C, no, mu_psi, parallel, cores
        )

        # geodesic paths
        q_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)
        f_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)

        for k in range(0, no):
            for l in range(0, Nstd):
                qhat = mqn + np.dot(U[0 : (M + 1), k], stds[l] * np.sqrt(s[k]))
                vechat = np.dot(U[(M + 1) :, k], (stds[l] * np.sqrt(s[k])) / C)
                psihat = geo.exp_map(mu_psi, vechat)
                gamhat = cumtrapz(psihat * psihat, np.linspace(0, 1, M), initial=0)
                gamhat = (gamhat - gamhat.min()) / (gamhat.max() - gamhat.min())
                if sum(vechat) == 0:
                    gamhat = np.linspace(0, 1, M)

                fhat = uf.cumtrapzmid(
                    time,
                    qhat[0:M] * np.fabs(qhat[0:M]),
                    np.sign(qhat[M]) * (qhat[M] * qhat[M]),
                    mididx,
                )
                f_pca[:, l, k] = uf.warp_f_gamma(np.linspace(0, 1, M), fhat, gamhat)
                q_pca[:, l, k] = uf.warp_q_gamma(
                    np.linspace(0, 1, M), qhat[0:M], gamhat
                )

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / s.sum()
            no = np.argwhere(cumm_coef >= var_exp)[0][0]

        self.q_pca = q_pca
        self.f_pca = f_pca
        self.latent = s[0:no]
        self.coef = a[:, 0:no]
        self.U = U[:, 0:no]
        self.mu_psi = mu_psi
        self.mu_g = mu_g
        self.id = mididx
        self.C = C
        self.time = time
        self.g = g
        self.cov = cov
        self.no = no
        self.stds = stds

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        fn = np.zeros((M, n))
        qn = np.zeros((M, n))
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])
            fn[:, ii] = fs.warp_f_gamma(self.time, f[:, ii], gam[:, ii])
            qn[:, ii] = fs.f_to_srsf(fn[:, ii], self.time)

        U = self.U
        no = U.shape[1]

        m_new = np.sign(fn[self.id, :]) * np.sqrt(np.abs(fn[self.id, :]))
        qn1 = np.vstack((qn, m_new))
        C = self.C
        TT = self.time.shape[0]
        mu_g = self.mu_g
        mu_psi = self.mu_psi
        vec = np.zeros((M, n))
        psi = np.zeros((TT, n))
        binsize = np.mean(np.diff(self.time))
        for i in range(0, n):
            psi[:, i] = np.sqrt(np.gradient(gam[:, i], binsize))
            out, theta = fs.inv_exp_map(mu_psi, psi[:, i])
            vec[:, i] = out

        g = np.vstack((qn1, C * vec))
        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                tmp = g[:, i] - mu_g
                a[i, j] = np.dot(tmp.T, U[:, j])

        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic joint fPCA result

        Usage: obj.plot()
        """
        no = self.no
        M = self.time.shape[0]
        Nstd = self.stds.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        cl = sorted(CBcdict.keys())
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(2, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[0, k1]
                if k > (no - 1):
                    break

                for l in range(0, Nstd):
                    axt.plot(self.time, self.q_pca[0:M, l, k], color=CBcdict[cl[l]])

                axt.set_title("q domain: PD %d" % (k + 1))

                axt = ax[1, k1]
                for l in range(0, Nstd):
                    axt.plot(self.time, self.f_pca[:, l, k], color=CBcdict[cl[l]])

                axt.set_title("f domain: PD %d" % (k + 1))

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent) / sum(self.latent)
        idx = np.arange(0, self.latent.shape[0]) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


def jointfPCAd(qn, vec, C, m, mu_psi, parallel, cores):
    (M, N) = qn.shape
    g = np.vstack((qn, C * vec))

    mu_q = qn.mean(axis=1)
    mu_g = g.mean(axis=1)

    K = np.cov(g)
    U, s, V = svd(K)

    a = np.zeros((N, m))
    for i in range(0, N):
        for j in range(0, m):
            tmp = g[:, i] - mu_g
            a[i, j] = np.dot(tmp.T, U[:, j])

    qhat = np.tile(mu_q, (N, 1))
    qhat = qhat.T
    qhat = qhat + np.dot(U[0:M, 0:m], a.T)

    vechat = np.dot(U[M:, 0:m], a.T / C)
    psihat = np.zeros((M - 1, N))
    gamhat = np.zeros((M - 1, N))
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(jfpca_sub)(mu_psi, vechat[:, n]) for n in range(N)
        )
        gamhat = np.array(out)
        gamhat = gamhat.transpose()
    else:
        for ii in range(0, N):
            psihat[:, ii] = geo.exp_map(mu_psi, vechat[:, ii])
            gam_tmp = cumtrapz(
                psihat[:, ii] * psihat[:, ii], np.linspace(0, 1, M - 1), initial=0
            )
            gamhat[:, ii] = (gam_tmp - gam_tmp.min()) / (gam_tmp.max() - gam_tmp.min())

    U = U[:, 0:m]
    s = s[0:m]

    return qhat, gamhat, a, U, s, mu_g, g, K


def jfpca_sub(mu_psi, vechat):
    M = mu_psi.shape[0]
    psihat = geo.exp_map(mu_psi, vechat)
    gam_tmp = cumtrapz(psihat * psihat, np.linspace(0, 1, M), initial=0)
    gamhat = (gam_tmp - gam_tmp.min()) / (gam_tmp.max() - gam_tmp.min())

    return gamhat


def find_C(C, qn, vec, q0, m, mu_psi, parallel, cores):
    qhat, gamhat, a, U, s, mu_g, g, K = jointfPCAd(
        qn, vec, C, m, mu_psi, parallel, cores
    )
    (M, N) = qn.shape
    time = np.linspace(0, 1, M - 1)

    d = np.zeros(N)
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(find_C_sub)(time, qhat[0 : (M - 1), n], gamhat[:, n], q0[:, n])
            for n in range(N)
        )
        d = np.array(out)
    else:
        for i in range(0, N):
            tmp = uf.warp_q_gamma(
                time, qhat[0 : (M - 1), i], uf.invertGamma(gamhat[:, i])
            )
            d[i] = trapz((tmp - q0[:, i]) * (tmp - q0[:, i]), time)

    out = sum(d * d) / N

    return out


def find_C_sub(time, qhat, gamhat, q0):
    tmp = uf.warp_q_gamma(time, qhat, uf.invertGamma(gamhat))
    d = trapz((tmp - q0) * (tmp - q0), time)

    return d



"""
functions for SRVF curve manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from scipy.interpolate import InterpolatedUnivariateSpline, interp1d
from scipy.integrate import trapz, cumtrapz
from numpy import zeros, ones, cumsum, linspace, gradient, sqrt
from numpy import eye, roll, tile, array, cos, sin, ascontiguousarray
from numpy import arccos, fabs, floor, fliplr, log, real, diff, mean
from scipy.linalg import norm, svd, det, solve
import optimum_reparam_N as orN
import fdasrsf.utility_functions as uf
from fdasrsf.rbfgs import rlbfgs


def resamplecurve(x, N=100, time=None, mode="O"):
    """
    This function resamples a curve to have N samples

    :param x: numpy ndarray of shape (2,M) of M samples
    :param N: Number of samples for new curve (default = 100)
    :param time: timing vector (Default=None)
    :param mode: Open ('O') or closed curve ('C') (default 'O')

    :rtype: numpy ndarray
    :return xn: resampled curve

    """
    n, T = x.shape
    xn = zeros((n, N))

    tst = x[:, 1] - x[:, 0]
    if tst[0] < 0:
        x = fliplr(x)

    delta = zeros(T)
    for r in range(1, T):
        delta[r] = norm(x[:, r] - x[:, r - 1])

    if time is None:
        cumdel = cumsum(delta) / delta.sum()
    else:
        time -= time[0]
        time /= time[-1]
        cumdel = time
    newdel = linspace(0, 1, N)

    for r in range(0, n):
        s = InterpolatedUnivariateSpline(cumdel, x[r, :], k=3)
        xn[r, :] = s(newdel)

    if mode == "C":
        q = curve_to_q(xn)[0]
        qn = project_curve(q)
        xn = q_to_curve(qn)

    return xn


def calculatecentroid(beta):
    """
    This function calculates centroid of a parameterized curve

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return centroid: center coordinates

    """
    n, T = beta.shape
    betadot = gradient(beta, 1.0 / (T - 1))
    betadot = betadot[1]
    normbetadot = zeros(T)
    integrand = zeros((n, T))
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])
        integrand[:, i] = beta[:, i] * normbetadot[i]

    scale = trapz(normbetadot, linspace(0, 1, T))
    centroid = trapz(integrand, linspace(0, 1, T), axis=1) / scale

    return centroid


def curve_to_q(beta, mode="O"):
    """
    This function converts curve beta to srvf q

    :param beta: numpy ndarray of shape (2,M) of M samples
    :param mode: Open ('O') or closed curve ('C') (default 'O')

    :rtype: numpy ndarray
    :return q: srvf of curve
    :return lenb: length of curve
    :return lenq: length of srvf

    """
    n, T = beta.shape
    v = gradient(beta, 1.0 / (T - 1))
    v = v[1]

    q = zeros((n, T))
    lenb = sqrt(innerprod_q2(v, v))
    for i in range(0, T):
        L = sqrt(norm(v[:, i]))
        if L > 0.0001:
            q[:, i] = v[:, i] / L
        else:
            q[:, i] = v[:, i] * 0.0001

    lenq = sqrt(innerprod_q2(q, q))
    q = q / lenq

    if mode == "C":
        q = project_curve(q)

    return (q, lenb, lenq)


def q_to_curve(q, scale=1):
    """
    This function converts srvf to beta

    :param q: numpy ndarray of shape (n,M) of M samples
    :param scale: scale of curve

    :rtype: numpy ndarray
    :return beta: parameterized curve

    """
    n, T = q.shape
    qnorm = zeros(T)
    for i in range(0, T):
        qnorm[i] = norm(q[:, i])

    beta = zeros((n, T))
    for i in range(0, n):
        beta[i, :] = cumtrapz(q[i, :] * qnorm, initial=0) / T

    beta = scale * beta

    return beta


def Basis_Normal_A(q):
    """
    Find Normal Basis

    :param q: numpy ndarray (n,T) defining T points on n dimensional SRVF

    :rtype list
    :return delg: basis
    """
    n, T = q.shape
    e = eye(n)
    Ev = zeros((n, T, n))
    for i in range(0, n):
        Ev[:, :, i] = tile(e[:, i], (T, 1)).T

    qnorm = zeros(T)
    for t in range(0, T):
        qnorm[t] = norm(q[:, t])

    delG = list()
    for i in range(0, n):
        tmp1 = tile(q[i, :] / qnorm, (n, 1))
        tmp2 = tile(qnorm, (n, 1))
        delG.append(tmp1 * q + tmp2 * Ev[:, :, i])

    return delG


def optimum_reparam_curve(q1, q2, lam=0.0, method="DP"):
    """
    calculates the warping to align srsf q2 to q1

    :param q1: matrix of size nxN or array of NxM samples of first SRVF
    :param time: vector of size N describing the sample points
    :param q2: matrix of size nxN or array of NxM samples samples of second
               SRVF
    :param lam: controls the amount of elasticity (default = 0.0)
    :param method: method to apply optimization (default="DP") options are
                   "DP" or "RBFGS"

    :rtype: vector
    :return gam: describing the warping function used to align q2 with q1

    """
    time = linspace(0, 1, q1.shape[1])
    if method == "DP":
        gam = orN.coptimum_reparam_curve(
            ascontiguousarray(q1), time, ascontiguousarray(q2), lam
        )
    elif method == "RBFGS":
        obj = rlbfgs(q1, q2, time)
        obj.solve()
        gam = obj.gammaOpt
    else:
        raise Exception("Invalid Optimization Method")

    return gam


def innerprod_q2(q1, q2):
    """
    This function calculates the inner product in srvf space

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return val: inner product

    """
    T = q1.shape[1]
    val = sum(sum(q1 * q2)) / T

    return val


def find_best_rotation(q1, q2, allow_reflection=False, only_xy=False):
    """
    This function calculates the best rotation between two srvfs using
    procustes rigid alignment

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param allow_reflection: bool indicating if reflection is allowed
                             (i.e. if the determinant of the optimal
                             rotation can be -1)
    :param only_xy: bool indicating if rotation should only be allowed
                    in the first two dimensions of the space

    :rtype: numpy ndarray
    :return q2new: optimal rotated q2 to q1
    :return R: rotation matrix

    """
    if q1.ndim != 2 or q2.ndim != 2:
        raise Exception(
            "This only supports curves of shape (N,M) for N dimensions and M samples"
        )

    n = q1.shape[0]

    # if only_xy, strip everything but the x and y coordinates of q1 and q2
    if only_xy:
        _q1 = q1[0:2, :]
        _q2 = q2[0:2, :]
    else:
        _q1 = q1
        _q2 = q2

    _n = _q1.shape[0]
    A = _q1 @ _q2.T
    U, s, Vh = svd(A)
    S = eye(_n)

    # if reflections are not allowed and the determinant of A is negative,
    # then the entry corresponding to the smallest singular value is negated
    # as in the Kabsch algorithm
    if det(A) < 0 and not allow_reflection:
        S[-1, -1] = -1  # the last entry of the matrix becomes -1

    _R = U @ S @ Vh  # optimal

    # if only_xy, the top left block of the matrix is _R and the rest is
    # identity matrix
    if only_xy:
        R = eye(n)
        R[0:2, 0:2] = _R
    else:
        R = _R

    q2new = R @ q2

    return (q2new, R)


def calculate_variance(beta):
    """
    This function calculates variance of curve beta

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return variance: variance

    """
    n, T = beta.shape
    betadot = gradient(beta, 1.0 / (T - 1))
    betadot = betadot[1]
    normbetadot = zeros(T)
    centroid = calculatecentroid(beta)
    integrand = zeros((n, n, T))
    t = linspace(0, 1, T)
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])
        a1 = beta[:, i] - centroid
        a1 = a1.reshape((n, 1))
        integrand[:, :, i] = a1 @ a1.T * normbetadot[i]

    l = trapz(normbetadot, t)
    variance = trapz(integrand, t, axis=2)
    variance /= l

    return variance


def psi(x, a, q):
    """
    This function formats variance output

    :param x: numpy ndarray of shape (2,M) of M samples curve
    :param a: numpy ndarray of shape (2,1) mean
    :param q: numpy ndarray of shape (2,M) of M samples srvf

    :rtype: numpy ndarray
    :return psi1: variance
    :return psi2: cross variance
    :return psi3: curve end
    :return psi4: curve end

    """
    T = q.shape[1]
    covmat = calculate_variance(x + tile(a, [T, 1]).T)
    psi1 = covmat[0, 0] - covmat[1, 1]
    psi2 = covmat[0, 1]
    psi3 = x[0, -1]
    psi4 = x[1, -1]

    return (psi1, psi2, psi3, psi4)


def find_basis_normal(q):
    """
    Finds the basis normal to the srvf

    :param q1: numpy ndarray of shape (2,M) of M samples

    :rtype: list of numpy ndarray
    :return basis: list containing basis vectors

    """
    n, T = q.shape

    f1 = zeros((n, T))
    f2 = zeros((n, T))
    for i in range(0, T):
        f1[:, i] = q[0, i] * q[:, i] / norm(q[:, i]) + array([norm(q[:, i]), 0])
        f2[:, i] = q[1, i] * q[:, i] / norm(q[:, i]) + array([0, norm(q[:, i])])

    h3 = f1
    h4 = f2
    integrandb3 = zeros(T)
    integrandb4 = zeros(T)
    for i in range(0, T):
        a = q[:, i].T
        integrandb3[i] = a.dot(h3[:, i])
        integrandb4[i] = a.dot(h4[:, i])

    b3 = h3 - q * trapz(integrandb3, linspace(0, 1, T))
    b4 = h4 - q * trapz(integrandb4, linspace(0, 1, T))

    basis = [b3, b4]

    return basis


def calc_j(basis):
    """
    Calculates Jacobian matrix from normal basis

    :param basis: list of numpy ndarray of shape (2,M) of M samples basis

    :rtype: numpy ndarray
    :return j: Jacobian

    """
    b1 = basis[0]
    b2 = basis[1]
    T = b1.shape[1]
    integrand11 = zeros(T)
    integrand12 = zeros(T)
    integrand22 = zeros(T)

    for i in range(0, T):
        a = b1[:, i].T
        b = b2[:, i].T
        integrand11[i] = a.dot(b1[:, i])
        integrand12[i] = a.dot(b2[:, i])
        integrand22[i] = b.dot(b2[:, i])

    j = zeros((2, 2))
    j[0, 0] = trapz(integrand11, linspace(0, 1, T))
    j[0, 1] = trapz(integrand12, linspace(0, 1, T))
    j[1, 1] = trapz(integrand22, linspace(0, 1, T))
    j[1, 0] = j[0, 1]

    return j


def shift_f(f, tau):
    """
    shifts a curve f by tau

    :param f: numpy ndarray of shape (2,M) of M samples
    :param tau: scalar

    :rtype: numpy ndarray
    :return fn: shifted curve

    """
    n, T = f.shape
    fn = zeros((n, T))
    fn[:, 0 : (T - 1)] = roll(f[:, 0 : (T - 1)], tau, axis=1)
    fn[:, T - 1] = fn[:, 0]

    return fn


def find_rotation_and_seed_unique(
    q1, q2, closed=0, lam=0.0, rotation=True, method="DP"
):
    """
    This function returns a candidate list of optimally oriented and
    registered (seed) shapes w.r.t. beta1

    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples
    :param closed: Open (0) or Closed (1)
    :param rotation: find rotation (default=True)
    :param lam: controls the elasticity (default = 0)
    :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"

    :rtype: numpy ndarray
    :return beta2new: optimal rotated beta2 to beta1
    :return O: rotation matrix
    :return tau: seed
    """

    n, T = q1.shape

    scl = 4.0
    minE = 1000
    if closed == 1:
        end_idx = int(floor(T / scl))
        scl = 4
    else:
        end_idx = 0

    for ctr in range(0, end_idx + 1):
        if closed == 1:
            q2n = shift_f(q2, scl * ctr)
        else:
            q2n = q2.copy()

        if rotation:
            q2new, R = find_best_rotation(q1, q2n)
        else:
            q2new = q2n
            R = eye(n)

        # Reparam
        if norm(q1 - q2new, "fro") > 0.0001:
            gam = optimum_reparam_curve(q2new, q1, lam, method)
            gamI = uf.invertGamma(gam)
            p2n = q_to_curve(q2n)
            p2n = group_action_by_gamma_coord(p2n, gamI)
            q2new = curve_to_q(p2n)[0]
            if closed == 1:
                q2new = project_curve(q2new)
        else:
            gamI = linspace(0, 1, T)

        tmp = innerprod_q2(q1, q2new)
        if tmp > 1:
            tmp = 1
        if tmp < -1:
            tmp = -1
        Ec = arccos(tmp)
        if Ec < minE:
            Rbest = R
            q2best = q2new
            gamIbest = gamI
            minE = Ec

    return (q2best, Rbest, gamIbest)


def find_rotation_and_seed_coord(beta1, beta2, closed=0, rotation=True, method="DP"):
    """
    This function returns a candidate list of optimally oriented and
    registered (seed) shapes w.r.t. beta1

    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples
    :param closed: Open (0) or Closed (1)
    :param rotation: find rotation (default=True)
    :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"

    :rtype: numpy ndarray
    :return beta2new: optimal aligned beta2 to beta1
    :return q2best: optimal aligned q2 to q1
    :return Rbest: rotation matrix
    :return gamIbest: warping function
    """

    n, T = beta1.shape
    q1 = curve_to_q(beta1)[0]
    scl = 4.0
    minE = 1000
    if closed == 1:
        end_idx = int(floor(T / scl))
        scl = 4
    else:
        end_idx = 0

    for ctr in range(0, end_idx + 1):
        if closed == 1:
            beta2n = shift_f(beta2, scl * ctr)
        else:
            beta2n = beta2

        if rotation:
            beta2new, R = find_best_rotation(beta1, beta2n)
        else:
            beta2new = beta2n
            R = eye(n)
        q2new = curve_to_q(beta2new)[0]

        # Reparam
        if norm(q1 - q2new, "fro") > 0.0001:
            gam = optimum_reparam_curve(q2new, q1, 0.0, method)
            gamI = uf.invertGamma(gam)
            beta2new = group_action_by_gamma_coord(beta2new, gamI)
            q2new = curve_to_q(beta2new)[0]
            if closed == 1:
                q2new = project_curve(q2new)
        else:
            gamI = linspace(0, 1, T)

        tmp = innerprod_q2(q1, q2new)
        if tmp > 1:
            tmp = 1
        if tmp < -1:
            tmp = -1
        Ec = arccos(tmp)
        if Ec < minE:
            Rbest = R
            beta2best = beta2new
            q2best = q2new
            gamIbest = gamI
            minE = Ec

    return (beta2best, q2best, Rbest, gamIbest)


def find_rotation_and_seed_q(q1, q2, closed=0, rotation=True, method="DP"):
    """
    This function returns a candidate list of optimally oriented and
    registered (seed) srvs w.r.t. q1

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param closed: Open (0) or Closed (1)
    :param rotation: find rotation (default=True)
    :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"

    :rtype: numpy ndarray
    :return q2best: optimal aligned q2 to q1
    :return Rbest: rotation matrix
    :return gamIbest: warping function
    """

    n, T = q1.shape
    scl = 4.0
    minE = 4000
    if closed == 1:
        end_idx = int(floor(T / scl))
        scl = 4
    else:
        end_idx = 0

    for ctr in range(0, end_idx + 1):
        if closed == 1:
            q2n = shift_f(q2, scl * ctr)
        else:
            q2n = q2

        if rotation:
            q2new, R = find_best_rotation(q1, q2n)
        else:
            q2new = q2n.copy()
            R = eye(n)

        # Reparam
        if norm(q1 - q2new, "fro") > 0.0001:
            gam = optimum_reparam_curve(q2new, q1, 0.0, method)
            gamI = uf.invertGamma(gam)
            q2new = group_action_by_gamma(q2new, gamI)
            if closed == 1:
                q2new = project_curve(q2new)
        else:
            gamI = linspace(0, 1, T)

        tmp = innerprod_q2(q1, q2new)
        if tmp > 1:
            tmp = 1
        if tmp < -1:
            tmp = -1
        Ec = arccos(tmp)
        if Ec < minE:
            Rbest = R
            q2best = q2new
            gamIbest = gamI
            minE = Ec

    return (q2best, Rbest, gamIbest)


def group_action_by_gamma_coord(f, gamma):
    """
    This function reparamerized curve f by gamma

    :param f: numpy ndarray of shape (2,M) of M samples
    :param gamma: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return fn: reparatermized curve

    """
    n, T = f.shape
    fn = zeros((n, T))

    for j in range(0, n):
        s = interp1d(linspace(0, 1, T), f[j, :])
        fn[j, :] = s(gamma)

    return fn


def group_action_by_gamma(q, gamma):
    """
    This function reparamerized srvf q by gamma

    :param f: numpy ndarray of shape (2,M) of M samples
    :param gamma: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return qn: reparatermized srvf

    """
    n, T = q.shape
    gammadot = gradient(gamma, 1.0 / T)
    qn = zeros((n, T))

    for j in range(0, n):
        s = InterpolatedUnivariateSpline(linspace(0, 1, T), q[j, :], k=3)
        qn[j, :] = s(gamma) * sqrt(gammadot)

    qn = qn / sqrt(innerprod_q2(qn, qn))

    return qn


def project_curve(q):
    """
    This function projects srvf q to set of close curves

    :param q: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return qproj: project srvf
    """
    n, T = q.shape
    if n == 2:
        dt = 0.35
    if n == 3:
        dt = 0.2
    epsilon = 1e-6

    iter = 1
    res = ones(n)
    J = zeros((n, n))

    s = linspace(0, 1, T)

    qnew = q.copy()
    qnew = qnew / sqrt(innerprod_q2(qnew, qnew))

    qnorm = zeros(T)
    G = zeros(n)
    C = zeros(300)
    while norm(res) > epsilon:
        if iter > 300:
            break

        # Jacobian
        for i in range(0, n):
            for j in range(0, n):
                J[i, j] = 3 * trapz(qnew[i, :] * qnew[j, :], s)

        J += eye(n)

        for i in range(0, T):
            qnorm[i] = norm(qnew[:, i])

        # Compute the residue
        for i in range(0, n):
            G[i] = trapz(qnew[i, :] * qnorm, s)

        res = -G

        if norm(res) < epsilon:
            break

        x = solve(J, res)
        C[iter] = norm(res)

        delG = Basis_Normal_A(qnew)
        temp = zeros((n, T))
        for i in range(0, n):
            temp += x[i] * delG[i] * dt

        qnew += temp
        iter += 1

    qnew = qnew / sqrt(innerprod_q2(qnew, qnew))

    return qnew


def pre_proc_curve(beta, T=100):
    """
    This function prepcoessed a curve beta to set of closed curves

    :param beta: numpy ndarray of shape (2,M) of M samples
    :param T: number of samples (default = 100)

    :rtype: numpy ndarray
    :return betanew: projected beta
    :return qnew: projected srvf
    :return A: alignment matrix (not used currently)

    """
    beta = resamplecurve(beta, T)
    q = curve_to_q(beta)[0]
    qnew = project_curve(q)
    x = q_to_curve(qnew)
    a = -1 * calculatecentroid(x)
    betanew = x + tile(a, [T, 1]).T
    A = eye(2)

    return (betanew, qnew, A)


def elastic_distance_curve(
    beta1, beta2, closed=0, rotation=True, scale=False, method="DP"
):
    """
    Calculates the two elastic distances between two curves in R^M
    :param beta1: numpy ndarray of shape (M,N) of N samples
    :param beta2: numpy ndarray of shape (M,N) of N samples
    :param closed: open (0) or closed (1) curve (default=0)
    :param rotation: compute optimal rotation (default=True)
    :param scale: include scale (default=False)
    :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"

    :rtype: tuple
    :return dist: shape distance
    :return dx: phase distance
    """

    if (beta1 == beta2).all():
        d = 0.0
        dx = 0.0
    else:
        N = beta1.shape[1]
        a = -calculatecentroid(beta1)
        beta1 += tile(a, (N, 1)).T
        a = -calculatecentroid(beta2)
        beta2 += tile(a, (N, 1)).T

        q1, len1, lenq1 = curve_to_q(beta1)
        q2, len2, lenq2 = curve_to_q(beta2)

        # compute shooting vector from q1 to q2
        beta2best, qn_t, Rbest, gam = find_rotation_and_seed_coord(
            beta1, beta2, closed, rotation, method
        )

        q1dotq2 = innerprod_q2(q1, qn_t)
        if q1dotq2 > 1:
            q1dotq2 = 1
        elif q1dotq2 < -1:
            q1dotq2 = -1

        if scale:
            d = sqrt(arccos(q1dotq2) ** 2 + log(lenq1 / lenq2) ** 2)
        else:
            d = arccos(q1dotq2)

        time1 = linspace(0, 1, N)
        binsize = mean(diff(time1))
        psi = sqrt(gradient(gam, binsize))
        q1dotq2 = trapz(psi, time1)
        if q1dotq2 > 1:
            q1dotq2 = 1
        elif q1dotq2 < -1:
            q1dotq2 = -1

        dx = real(arccos(q1dotq2))

    return d, dx


def inverse_exp_coord(beta1, beta2, closed=0, method="DP"):
    """
    Calculate the inverse exponential to obtain a shooting vector from
    beta1 to beta2 in shape space of open curves

    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples
    :param closed: open (0) or closed (1) curve
    :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"

    :rtype: numpy ndarray
    :return v: shooting vectors
    :return dist: distance

    """
    T = beta1.shape[1]
    centroid1 = calculatecentroid(beta1)
    beta1 = beta1 - tile(centroid1, [T, 1]).T
    centroid2 = calculatecentroid(beta2)
    beta2 = beta2 - tile(centroid2, [T, 1]).T

    q1 = curve_to_q(beta1)[0]

    # Iteratively optimize over SO(n) x Gamma
    beta2n, q2n, O_hat, gamI = find_rotation_and_seed_coord(
        beta1, beta2, closed, method
    )

    # Compute geodesic distance
    q1dotq2 = innerprod_q2(q1, q2n)
    if q1dotq2 > 1:
        q1dotq2 = 1
    elif q1dotq2 < -1:
        q1dotq2 = -1
    dist = arccos(q1dotq2)

    # Compute shooting vector
    if q1dotq2 > 1:
        q1dotq2 = 1

    u = q2n - q1dotq2 * q1
    normu = sqrt(innerprod_q2(u, u))

    if normu > 1e-4:
        v = u * arccos(q1dotq2) / normu
    else:
        v = zeros((2, T))

    return (v, dist)


def inverse_exp(q1, q2, beta2):
    """
    Calculate the inverse exponential to obtain a shooting vector from
    q1 to q2 in shape space of open curves

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return v: shooting vectors

    """
    T = q1.shape[1]
    centroid1 = calculatecentroid(beta2)
    beta2 = beta2 - tile(centroid1, [T, 1]).T

    # Optimize over SO(n)
    q2, O_hat, gamI = find_rotation_and_seed_q(q1, q2)

    # Applying optimal re-parameterization to the second curve
    beta2 = group_action_by_gamma_coord(beta2, gamI)
    q2 = curve_to_q(beta2)

    # Compute geodesic distance
    q1dotq2 = innerprod_q2(q1, q2)
    dist = arccos(q1dotq2)

    # Compute shooting vector
    if q1dotq2 > 1:
        q1dotq2 = 1

    u = q2 - q1dotq2 * q1
    normu = sqrt(innerprod_q2(u, u))

    if normu > 1e-4:
        v = u * arccos(q1dotq2) / normu
    else:
        v = zeros((2, T))

    return v


def gram_schmidt(basis):
    """
    Performs Gram Schmidt Orthogonlization of a basis_o

     :param basis: list of numpy ndarray of shape (2,M) of M samples

     :rtype: list of numpy ndarray
     :return basis_o: orthogonlized basis

    """
    b1 = basis[0]
    b2 = basis[1]

    basis1 = b1 / sqrt(innerprod_q2(b1, b1))
    b2 = b2 - innerprod_q2(basis1, b2) * basis1
    basis2 = b2 / sqrt(innerprod_q2(b2, b2))

    basis_o = [basis1, basis2]

    return basis_o


def project_tangent(w, q, basis):
    """
    projects srvf to tangent space w using basis

    :param w: numpy ndarray of shape (2,M) of M samples
    :param q: numpy ndarray of shape (2,M) of M samples
    :param basis: list of numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return wproj: projected q

    """
    w = w - innerprod_q2(w, q) * q
    bo = gram_schmidt(basis)

    wproj = w - innerprod_q2(w, bo[0]) * bo[0] - innerprod_q2(w, bo[1]) * bo[1]

    return wproj


def scale_curve(beta):
    """
    scales curve to length 1

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return beta_scaled: scaled curve
    :return scale: scale factor used

    """
    n, T = beta.shape
    normbetadot = zeros(T)
    betadot = gradient(beta, 1.0 / T)
    betadot = betadot[1]
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])

    scale = trapz(normbetadot, linspace(0, 1, T))
    beta_scaled = beta / scale

    return (beta_scaled, scale)


def parallel_translate(w, q1, q2, basis, mode=0):
    """
    parallel translates q1 and q2 along manifold

    :param w: numpy ndarray of shape (2,M) of M samples
    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param basis: list of numpy ndarray of shape (2,M) of M samples
    :param mode: open 0 or closed curves 1 (default 0)

    :rtype: numpy ndarray
    :return wbar: translated vector

    """
    modes = [0, 1]
    mode = [i for i, x in enumerate(modes) if x == mode]
    if len(mode) == 0:
        mode = 0
    else:
        mode = mode[0]

    wtilde = w - 2 * innerprod_q2(w, q2) / innerprod_q2(q1 + q2, q1 + q2) * (q1 + q2)
    l = sqrt(innerprod_q2(wtilde, wtilde))

    if mode == 1:
        wbar = project_tangent(wtilde, q2, basis)
        normwbar = sqrt(innerprod_q2(wbar, wbar))
        if normwbar > 10 ** (-4):
            wbar = wbar * l / normwbar
    else:
        wbar = wtilde

    return wbar


def curve_zero_crossing(Y, q, bt, y_max, y_min, gmax, gmin):
    """
    finds zero-crossing of optimal gamma, gam = s*gmax + (1-s)*gmin
    from elastic curve regression model

    :param Y: response
    :param beta: predicitve function
    :param bt: basis function
    :param y_max: maximum repsonse for warping function gmax
    :param y_min: minimum response for warping function gmin
    :param gmax: max warping function
    :param gmin: min warping fucntion

    :rtype: numpy array
    :return gamma: optimal warping function
    :return O_hat: rotation matrix

    """
    # simple iterative method based on intermediate theorem
    n = q.shape[0]
    T = q.shape[1]
    max_itr = 100
    a = zeros(max_itr)
    a[0] = 1
    f = zeros(max_itr)
    f[0] = y_max - Y
    f[1] = y_min - Y
    mrp = f[0]
    mrn = f[1]
    mrp_ind = 0  # most recent positive index
    mrn_ind = 1  # most recent negative index

    for ii in range(2, max_itr):
        x1 = a[mrp_ind]
        x2 = a[mrn_ind]
        y1 = mrp
        y2 = mrn
        a[ii] = (x1 * y2 - x2 * y1) / float(y2 - y1)

        gamma = a[ii] * gmax + (1 - a[ii]) * gmin

        q1 = group_action_by_gamma(q, gamma)
        # q1, O_hat1 = find_best_rotation(bt, q1)
        f[ii] = innerprod_q2(q1, bt) - Y

        if fabs(f[ii]) < 1e-5:
            break
        elif f[ii] > 0:
            mrp = f[ii]
            mrp_ind = ii
        else:
            mrn = f[ii]
            mrn_ind = ii

    gamma = a[ii] * gmax + (1 - a[ii]) * gmin

    # q1 = group_action_by_gamma(q, gamma)
    # q1, O_hat = find_best_rotation(bt, q1)
    O_hat = eye(n)

    return (gamma, O_hat)


def elastic_shooting(q1, v, mode=0):
    """
    Calculates shooting vector from v to q1

    :param q1: vector of srvf
    :param v: shooting vector
    :param mode: closed or open (1/0)

    :rtype numpy ndarray
    :return q2n: vector of srvf
    """
    d = sqrt(innerprod_q2(v, v))
    if d < 0.00001:
        q2n = q1
    else:
        q2n = cos(d) * q1 + (sin(d) / d) * v
        if mode == 1:
            q2n = project_curve(q2n)

    return q2n


def elastic_shooting_vector(q1, q2, mode=0):
    """
    Calculates shooting between two srvfs

    :param q1: vector of srvf
    :param q2: vector of srvf
    :param mode: closed or open (1/0)

    :rtype numpy ndarray
    :return v: shooting vector
    :return d: distance
    :return q2n: aligned srvf
    """

    (q2n, Rbest, gamIbest) = find_rotation_and_seed_unique(q1, q2, closed=mode)
    lenq = sqrt(innerprod_q2(q2n, q2n))
    q2n = q2n / lenq

    d = sqrt(innerprod_q2(q1, q2n))
    if d < 0.00001:
        v = zeros(q1.shape)
    else:
        v = (d / sin(d)) * (q2n - cos(d) * q1)
        q2n = cos(d) * q1 + (sin(d) / d) * v
        if mode == 1:
            q2n = project_curve(q2n)

    return (v, d, q2n)


def rot_mat(theta):
    O = array([(cos(theta), -1 * sin(theta)), (sin(theta), cos(theta))])

    return O



"""
Elastic Functional Clustering

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf.utility_functions as uf
from scipy.integrate import trapz
from numpy.linalg import norm
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors


def kmeans_align(
    f,
    time,
    K,
    seeds=None,
    lam=0,
    showplot=True,
    smooth_data=False,
    parallel=False,
    alignment=True,
    omethod="DP2",
    MaxItr=50,
    thresh=0.01,
):
    """
    This function clusters functions and aligns using the elastic square-root
    slope (srsf) framework.

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param K: number of clusters
    :param seeds: indexes of cluster center functions (default = None)
    :param lam: controls the elasticity (default = 0)
    :param showplot: shows plots of functions (default = T)
    :param smooth_data: smooth data using box filter (default = F)
    :param parallel: enable parallel mode using \code{\link{joblib}} and \code{doParallel} package (default=F)
    :param alignment: whether to perform alignment (default = T)
    :param omethod: optimization method (DP,DP2,RBFGS)
    :param MaxItr: maximum number of iterations
    :param thresh: cost function threshold
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: dictionary
    :return fn: aligned functions - matrix (N x M) of M functions with N samples which is a list for each cluster
    :return qn: aligned SRSFs - similar structure to fn
    :return q0: original SRSFs
    :return labels: cluster labels
    :return templates: cluster center functions
    :return templates_q: cluster center SRSFs
    :return gam: warping functions - similar structure to fn
    :return qun: Cost Function

    """

    w = 0.0
    k = 1
    cores = -1
    eps = np.finfo(np.double).eps

    M = f.shape[0]
    N = f.shape[1]

    if seeds is None:
        a = np.arange(0, N, dtype=int)
        template_ind = np.random.choice(a, K)
    else:
        template_ind = seeds

    templates = np.zeros((M, K))
    for i in range(K):
        templates[:, i] = f[:, template_ind[i]]

    cluster_id = np.zeros(N, dtype=int)
    qun = np.zeros(MaxItr)

    # convert to SRSF
    f, g, g2 = uf.gradient_spline(time, f, smooth_data)
    q = g / np.sqrt(abs(g) + eps)
    templates_q = np.zeros((M, K))
    for i in range(K):
        templates_q[:, i] = q[:, template_ind[i]]

    for itr in range(0, MaxItr):
        print("updating step: r=%d" % (itr + 1))

        # Alignment
        gam = {}
        Dy = np.zeros((K, N))
        qn = {}
        fn = {}

        for k in range(K):
            gam_tmp = np.zeros((M, N))
            if alignment:
                if parallel:
                    out = Parallel(n_jobs=cores)(
                        delayed(uf.optimum_reparam)(
                            templates_q[:, k], time, q[:, n], omethod, lam
                        )
                        for n in range(N)
                    )
                    gam_tmp = np.array(out)
                    gam_tmp = gam_tmp.transpose()
                else:
                    for n in range(0, N):
                        gam_tmp[:, n] = uf.optimum_reparam(
                            templates_q[:, k], time, q[:, n], omethod, lam
                        )
            else:
                for n in range(0, N):
                    gam_tmp[:, k] = np.linspace(0, 1, M)

            fw = np.zeros((M, N))
            qw = np.zeros((M, N))
            dist = np.zeros(N)
            for i in range(0, N):
                fw[:, i] = uf.warp_f_gamma(time, f[:, i], gam_tmp[:, i])
                qw[:, i] = uf.f_to_srsf(fw[:, i], time)
                dist[i] = np.sqrt(trapz((qw[:, i] - templates_q[:, k]) ** 2, time))

            Dy[k, :] = dist
            qn[k] = qw
            fn[k] = fw
            gam[k] = gam_tmp

        # Assignment
        cluster_id = Dy.argmin(axis=0)

        # Normalization
        for k in range(K):
            idx = np.where(cluster_id == k)[0]
            ftmp = fn[k][:, idx]
            gamtmp = gam[k][:, idx]
            gamI = uf.SqrtMeanInverse(gamtmp)
            N1 = idx.shape[0]

            gamt = np.zeros((M, N1))
            f_temp = np.zeros((M, N1))
            q_temp = np.zeros((M, N1))
            if parallel:
                out = Parallel(n_jobs=cores)(
                    delayed(norm_sub)(ftmp[:, i], time, gamtmp[:, i], gamI)
                    for i in range(N1)
                )
                for i in range(0, N1):
                    f_temp[:, i] = out[i][0]
                    q_temp[:, i] = out[i][1]
                    gamt[:, i] = out[i][2]
            else:
                for i in range(N1):
                    f_temp[:, i], q_temp[:, i], gamt[:, i] = norm_sub(
                        ftmp[:, i], time, gamtmp[:, i], gamI
                    )

            qn[k][:, idx] = q_temp
            fn[k][:, idx] = f_temp
            gam[k][:, idx] = gamt

        # Template Identification
        qun_t = np.zeros(K)
        old_templates_q = templates_q.copy()
        for k in range(K):
            idx = np.where(cluster_id == k)[0]
            templates_q[:, k] = qn[k][:, idx].mean(axis=1)
            templates[:, k] = fn[k][:, idx].mean(axis=1)

            qun_t[k] = norm(templates_q[:, k] - old_templates_q[:, k]) / norm(
                old_templates_q[:, k]
            )

        qun[itr] = qun_t.mean()

        if qun[itr] < thresh:
            break

    # Output
    ftmp = {}
    qtmp = {}
    gamtmp = {}
    for k in range(K):
        idx = np.where(cluster_id == k)[0]
        ftmp[k] = fn[k][:, idx]
        qtmp[k] = qn[k][:, idx]
        gamtmp[k] = gam[k][:, idx]

    out = {}
    out["f0"] = f
    out["q0"] = q
    out["time"] = time
    out["fn"] = ftmp
    out["qn"] = qtmp
    out["gam"] = gamtmp
    out["labels"] = cluster_id
    out["templates"] = templates
    out["templates_q"] = templates_q
    out["lambda"] = lam
    out["omethod"] = omethod
    out["qun"] = qun[0:itr]

    if showplot:
        num_plot = int(np.ceil(K / 6))
        a = mcolors.TABLEAU_COLORS
        colors = list(a.keys())
        plt.figure()
        plt.plot(time, f)
        plt.title("Original Data")

        plt.figure()
        plt.plot(time, templates)
        plt.title("Cluster Mean Functions")

        for k in range(num_plot):
            cnt = 1
            plt.figure()
            for n in np.arange(k * 6, min(K, (k + 1) * 6), dtype=int):
                ax = plt.subplot(2, 3, cnt)
                ax.plot(time, ftmp[n], color="lightgrey")
                ax.plot(time, templates[:, n], color=colors[cnt - 1])
                ax.set_title("Cluster f: %d" % n)
                cnt += 1

        for k in range(num_plot):
            cnt = 1
            plt.figure()
            for n in np.arange(k * 6, min(K, (k + 1) * 6), dtype=int):
                ax = plt.subplot(2, 3, cnt)
                ax.plot(time, qtmp[n], color="lightgrey")
                ax.plot(time, templates_q[:, n], color=colors[cnt - 1])
                ax.set_title("Cluster q: %d" % n)
                cnt += 1

        plt.show()

    return out


def norm_sub(f, time, gam, gamI):
    fw = uf.warp_f_gamma(time, f, gamI)
    qw = uf.f_to_srsf(fw, time)
    time0 = (time[-1] - time[0]) * gamI + time[0]
    gamw = np.interp(time0, time, gam)

    return (fw, qw, gamw)



"""
image warping using SRVF framework

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf.image_functions as fif
import cimage as im


def reparam_image(It, Im, gam=None, b=None, stepsize=1e-4, itermax=20):
    """
    This function warps an image to another using SRVF framework

    :param Im: numpy ndarray of shape (N,N) representing a NxN image
    :param Im: numpy ndarray of shape (N,N) representing a NxN image
    :param gam: numpy ndarray of shape (N,N) representing an initial warping function
    :param b: numpy ndarray representing basis matrix

    :rtype: numpy ndarray
    :return gamnew: diffeomorphism
    :return Inew: warped image
    :return H: energy
    :return stepsize: final stepsize

    """

    m = It.shape[0]
    n = It.shape[1]

    gamid = fif.makediffeoid(m, n)

    if gam is None:
        gam = gamid.copy()

    if b is None:
        M = 10
        basetype = "t"
        b = fif.formbasisTid(M, m, n, basetype)

    # main loop
    H = np.zeros(itermax + 1)
    Iold = fif.apply_gam_imag(Im, gam)
    Iold -= Iold.min()
    Iold /= Iold.max()
    qt = fif.image_to_q(It)
    qm = fif.image_to_q(Iold)
    gamold = gam.copy()

    gamnew = gamold.copy()
    Inew = Iold.copy()
    iter = 0
    H[iter] = fif.compEnergy(qt, qm)

    print("Iteration %d, energy %f\n" % (iter, H[iter]))

    gamupdate = fif.updateGam(qt, qm, b)

    cutoff = 1e-3

    for iter in range(1, itermax + 1):
        gaminc = gamid + stepsize * gamupdate

        G = im.check_crossing(gamnew)

        if G == 0:
            print("Possible crossing!\n")
            gamnew = gamold.copy()
            stepsize *= 0.67
            H[iter] = H[iter - 1]
            continue

        else:
            gamnew = fif.apply_gam_gamid(gamnew, gaminc)

        Inew = fif.apply_gam_imag(Im, gamnew)
        Inew -= Inew.min()
        Inew /= Inew.max()
        qm = fif.image_to_q(Inew)
        H[iter] = fif.compEnergy(qt, qm)
        print("Iteration %d, energy %f\n" % (iter, H[iter]))

        if iter > 4:
            hstop = 1
            for i in range(4):
                hstop *= H[iter] >= H[iter - i]

            if hstop != 0:
                print("Warning: energy constantly increasing\n")
                break

        if (iter > 4) and (
            np.abs(H[iter] <= H[iter - 1]) < cutoff
            and np.abs(H[iter - 1] - H[iter - 2]) < cutoff
            and np.abs(H[iter - 2] <= H[iter - 3]) < cutoff
        ):
            print("Warning: energy is not changing\n")
            break

        if ((iter > 1) and (H[iter] > H[iter - 1])) or (
            (iter > 3)
            and ((H[iter - 1] <= H[iter - 2]) and (H[iter - 2] > H[iter - 3]))
        ):
            stepsize *= 0.9
            gamnew = gamold.copy()
            H[iter] = H[iter - 1]
            continue

        gamold = gamnew.copy()
        gamupdate = fif.updateGam(qt, qm, b)

    H = H[0:iter]

    return gamnew, Inew, H, stepsize



"""
Partial Least Squares using SVD

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
from fdasrsf.utility_functions import diffop, geigen, innerprod_q


def pls_svd(time, qf, qg, no, alpha=0.0):
    """
    This function computes the partial least squares using SVD

    :param time: vector describing time samples
    :param qf: numpy ndarray of shape (M,N) of N functions with M samples
    :param qg: numpy ndarray of shape (M,N) of N functions with M samples
    :param no: number of components
    :param alpha: amount of smoothing (Default = 0.0 i.e., none)

    :rtype: numpy ndarray
    :return wqf: f weight function
    :return wqg: g weight function
    :return alpha: smoothing value
    :return values: singular values

    """
    binsize = np.diff(time)
    binsize = binsize.mean()

    Kfg = np.cov(qf, qg)
    Kfg = Kfg[0 : qf.shape[0], qf.shape[0] : Kfg.shape[0]]
    nx = Kfg.shape[0]
    D4x = diffop(nx, binsize)
    values, Lmat, Mmat = geigen(Kfg, np.eye(nx) + alpha * D4x, np.eye(nx) + alpha * D4x)
    wf = Lmat[:, 0:no]
    wg = Mmat[:, 0:no]

    for ii in range(0, no):
        wf[:, ii] = wf[:, ii] / np.sqrt(innerprod_q(time, wf[:, ii], wf[:, ii]))
        wg[:, ii] = wg[:, ii] / np.sqrt(innerprod_q(time, wg[:, ii], wg[:, ii]))

    wqf = wf
    wqg = wg

    N = qf.shape[1]
    rfi = np.zeros(N)
    rgi = np.zeros(N)

    for l in range(0, N):
        rfi[l] = innerprod_q(time, qf[:, l], wqf[:, 0])
        rgi[l] = innerprod_q(time, qg[:, l], wqg[:, 0])

    cost = np.cov(rfi, rgi)[1, 0]

    return wqf, wqg, alpha, values, cost
